# -*- coding: utf-8 -*-

import os

from simmate.toolkit import Structure
from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.outputs import Oszicar


class Potim(ErrorHandler):
    """
    This handler checks if a run has excessively large positive energy changes,
    where it's typically caused by too large of a POTIM. Runs will end up crashing
    with some other error (e.g. BRMIX) as the geometry gets progressively worse.
    """

    is_monitor = True

    def __init__(self, dE_per_atom_threshold: float = 1):
        self.dE_per_atom_threshold = dE_per_atom_threshold

    def check(self, directory: str) -> bool:

        # We check for this error in the OSZICAR because it's the smallest file
        # that will tell us energies -- and therefore the fastest to read.
        oszicar_filename = os.path.join(directory, "OSZICAR")

        # check to see that the file is there first
        if os.path.exists(oszicar_filename):

            # then load the file's data
            oszicar = Oszicar(oszicar_filename)

            # also load the structure so we know how many sites there are
            poscar_filename = os.path.join(directory, "POSCAR")
            structure = Structure.from_file(poscar_filename)
            nsites = structure.num_sites

            # iterate through all of the ionic steps and look at the changes
            # in energy. If any is greater than our threshold, then we have
            # an error!
            # We skip the first ionic step too bc there is no energy change there.
            for ionic_step in oszicar.ionic_steps[1:]:
                energy_change_per_atom = ionic_step["energy_change"] / nsites
                if energy_change_per_atom > self.dE_per_atom_threshold:
                    return True

        # if the file doesn't exist OR the threshold is never hit, the we are
        # not seeing any error.
        return False

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        current_potim = incar.get("POTIM", 0.5)
        current_ibrion = incar.get("IBRION", 0)
        current_symprec = incar.get("SYMPREC", 1e-5)

        # if we have a small potim and an ibrion that isn't damped molecular
        # dynamics, then then we switch the damped MD
        if current_potim < 0.2 and current_ibrion != 3:
            incar["IBRION"] = 3
            incar["SMASS"] = 0.75
            correction = "switched IBRION to 3 and SMASS to 0.75"

        # If we already have a small POTIM, then we try setting a low SYMPREC
        elif current_potim < 0.1 and current_symprec != 1e-8:
            incar["SYMPREC"] = 1e-8
            correction = "switched SYMPREC to 1e-8"

        # Otherwise we trying halving the potim
        else:
            new_potim = current_potim * 0.5
            incar["POTIM"] = new_potim
            correction = f"halved the POTIM from {current_potim} to {new_potim}"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

import importlib.metadata

__version__ = importlib.metadata.version("simmate")

from simmate.utilities import get_doc_from_readme, check_if_using_latest_version

__doc__ = get_doc_from_readme(__file__)


# Check if there is a newer Simmate version available, and if the current
# installation is out of date, print a message for the user.
try:
    check_if_using_latest_version()
except:
    print("WARNING: Unable to check if using the latest Simmate version.")



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class ElfKpar(ErrorHandler):

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["ELF: KPAR>1 not implemented"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # make the fix
        incar["KPAR"] = 1
        correction = "switched KPAR to 1"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from .bader import PopulationAnalysis__Vasp__BaderMatproj
from .elf import PopulationAnalysis__Vasp__ElfMatproj
from .badelf import PopulationAnalysis__Vasp__BadelfMatproj

# NOTE: we do not import the independent pre-* workflows as these should not
# be called directly.



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from pymatgen.analysis.diffusion.neb.pathfinder import DistinctPathFinder
from .migration_hop import MigrationHop
from .migration_images import MigrationImages



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------

"""
Feel free to write your own StopCondition! It should follow the format shown here.

-Jack
"""


class StopCondition:
    def __init__(
        self,
    ):
        # add any unchanging variables or settings here
        pass

    def check(self, search):
        # The select method should take in one argument:
        # search = this is the main Search object seen in search.py
        # The StopCondition can run any analysis on the Search object WITHOUT
        # making any changes to it. Then you should return True if the
        # stopping condition(s) has been met and False if the calculation should
        # continue.
        pass


# -----------------------------------------------------------------------------


class BasicStopConditions:
    def __init__(self, max_structures=300, energy_limit=-999, same_min_structures=50):
        self.max_structures = max_structures
        self.energy_limit = energy_limit
        self.same_min_structures = same_min_structures

    def check(self, search):

        # Note, a single condition met is enough to stop the search (returns True)

        # start by looking at the total number of structures that completed their analysis
        if search.njobs_completed >= self.max_structures:
            print("\nMax structures hit. Stopping Search")
            return True

        # grab the completed energies and the lowest energy for reference
        energies = [e for e in search.fitnesses if e]  # remove None types
        if not energies:  # this will happen if the no structures have completed yet
            return False
        min_energy = min(energies)

        # look at the structure energies and see what the lowest is
        if min_energy <= self.energy_limit:
            print("\nEnergy limit hit. Stopping Search")
            return True

        # look how long the minimum structure has been there #!!! ERROR IN LOGIC
        # min_index = energies.index(min_energy)
        # if len(energies) - min_index >= self.same_min_structures:
        #     print('\nMin-structure reign limit hit. Stopping Search')
        #     return True

        # If no conditions above are met, return False to continue the search
        return False


# -----------------------------------------------------------------------------



# -*- coding: utf-8 -*-

from django.shortcuts import render

from simmate.database import third_parties
from simmate.database.base_data_types import DatabaseTable
from simmate.website.core_components.base_api_view import SimmateAPIViewSet


def providers_all(request):

    # TODO: auto determine this list and descriptions
    workflows_metadata = {
        # "AflowStructure": "...",  # not allowed to redistribute
        "AflowPrototype": "...",
        "CodStructure": "...",
        "JarvisStructure": "...",
        "MatprojStructure": "...",
        "OqmdStructure": "...",
    }

    # now let's put the data and template together to send the user
    context = {"workflows_metadata": workflows_metadata}
    template = "third_parties/providers_all.html"
    return render(request, template, context)


class ProviderAPIViewSet(SimmateAPIViewSet):

    template_list = "third_parties/provider.html"
    template_retrieve = "third_parties/entry_detail.html"

    @classmethod
    def get_table(
        cls,
        request,
        provider_name,
        pk=None,
    ) -> DatabaseTable:
        """
        grabs the relevant database table using the URL request
        """
        # using the provider name (which is really just the table name), load
        # the corresponding database table
        provider_table = getattr(third_parties, provider_name)
        return provider_table

    def get_list_context(
        self,
        request,
        provider_name,
    ) -> dict:

        provider_table = getattr(third_parties, provider_name)
        return {"provider": provider_table}

    def get_retrieve_context(
        self,
        request,
        provider_name,
        pk,
    ) -> dict:

        return {
            "provider_name": provider_name,
            "entry_id": pk,
        }



# -*- coding: utf-8 -*-

import os

from simmate.database.base_data_types import (
    table_column,
    Structure,
    Forces,
    Thermodynamics,
    Calculation,
)

from pymatgen.io.vasp.outputs import Vasprun


class StaticEnergy(Structure, Thermodynamics, Forces, Calculation):
    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = (
        [
            "valence_band_maximum",
            "conduction_band_minimum",
            "energy_fermi",
            "is_gap_direct",
        ]
        + Structure.base_info
        + Thermodynamics.base_info
        + Forces.base_info
        + Calculation.base_info
    )

    # OPTIMIZE: should I include this electronic data?

    band_gap = table_column.FloatField(blank=True, null=True)
    """
    The band gap energy in eV.
    """

    is_gap_direct = table_column.BooleanField(blank=True, null=True)
    """
    Whether the band gap is direct or indirect.
    """

    energy_fermi = table_column.FloatField(blank=True, null=True)
    """
    The Fermi energy in eV.
    """

    conduction_band_minimum = table_column.FloatField(blank=True, null=True)
    """
    The conduction band minimum in eV.
    """

    valence_band_maximum = table_column.FloatField(blank=True, null=True)
    """
    The valence band maximum in eV.
    """

    @classmethod
    def from_directory(cls, directory: str):
        # I assume the directory is from a vasp calculation, but I need to update
        # this when I begin adding new calculators.
        vasprun_filename = os.path.join(directory, "vasprun.xml")
        vasprun = Vasprun(vasprun_filename)
        return cls.from_vasp_run(vasprun)

    @classmethod
    def from_vasp_run(cls, vasprun: Vasprun):
        # Takes a pymatgen VaspRun object, which is what's typically returned
        # from a simmate VaspTask.run() call.

        # The data is actually easier to access as a dictionary and everything
        # we need is stored under the "output" key.
        data = vasprun.as_dict()["output"]
        # In a static energy calculation, there is only one ionic step so we
        # grab that up front.
        ionic_step = data["ionic_steps"][0]

        # Take our structure, energy, and forces to build all of our other
        # fields for this datatable
        static_energy = cls.from_toolkit(
            structure=vasprun.final_structure,
            energy=ionic_step["e_wo_entrp"],
            site_forces=ionic_step["forces"],
            lattice_stress=ionic_step["stress"],
            band_gap=data.get("bandgap"),
            is_gap_direct=data.get("is_gap_direct"),
            energy_fermi=data.get("efermi"),
            conduction_band_minimum=data.get("cbm"),
            valence_band_maximum=data.get("vbm"),
        )
        static_energy.save()
        return static_energy

    def update_from_vasp_run(
        self,
        vasprun: Vasprun,
        corrections: list,
        directory: str,
    ):
        # Takes a pymatgen VaspRun object, which is what's typically returned
        # from a simmate VaspTask.run() call.

        # The data is actually easier to access as a dictionary and everything
        # we need is stored under the "output" key.
        data = vasprun.as_dict()["output"]
        # In a static energy calculation, there is only one ionic step so we
        # grab that up front.
        ionic_step = data["ionic_steps"][0]

        # Take our structure, energy, and forces to build all of our other
        # fields for this datatable
        # OPTIMIZE: this overwrites structure data, which should already be there.
        # Is there a faster way to grab this data and update attributes?
        new_kwargs = self.from_toolkit(
            structure=vasprun.final_structure,
            energy=ionic_step["e_wo_entrp"],
            site_forces=ionic_step["forces"],
            lattice_stress=ionic_step["stress"],
            as_dict=True,
        )
        for key, value in new_kwargs.items():
            setattr(self, key, value)

        # There is also extra data for the final structure that we save directly
        # in the relaxation table. We use .get() in case the key isn't provided
        self.band_gap = data.get("bandgap")
        self.is_gap_direct = data.get("is_gap_direct")
        self.energy_fermi = data.get("efermi")
        self.conduction_band_minimum = data.get("cbm")
        self.valence_band_maximum = data.get("vbm")

        # lastly, we also want to save the corrections made and directory it ran in
        self.corrections = corrections
        self.directory = directory

        # Now we have the relaxation data all loaded and can save it to the database
        self.save()



# -*- coding: utf-8 -*-

import os

from simmate.toolkit import Structure
from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.outputs import Oszicar


class Zpotrf(ErrorHandler):
    """
    This handler attempts to fix a routine failure associated with LAPACK. The
    fix depends on the state of the calculation, so we check the OSZICAR to
    decide what to do.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["LAPACK: Routine ZPOTRF failed"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # We can learn more about the cause of this error by looking at
        # the OSZICAR and checking the number of steps completed.
        # BUG: what if the OSCZICAR is improperly formatted but a number
        # of ionic steps completed successfully?
        try:
            oszicar_filename = os.path.join(directory, "OSZICAR")
            oszicar = Oszicar(oszicar_filename)
            nionic_steps = len(oszicar.ionic_steps)
        except Exception:
            nionic_steps = 0

        # If the error shows up after the first ionic step, then this is likely
        # caused by POTIM being too large.
        if nionic_steps >= 1:
            # Halve the POTIM and turn off symmetry
            current_potim = incar.get("POTIM", 0.5)
            new_potim = current_potim / 2
            incar["POTIM"] = new_potim
            incar["ISYM"] = 0
            correction = (
                f"set ISYM to 0 and switched POTIM from {current_potim} to {new_potim}"
            )

        # If this is a static calculation (NSW=0) or if ISIF is 1-3, then
        # we just want to turn off symmetry
        elif incar.get("NSW", 0) == 0 or incar.get("ISIF", 0) in [0, 1, 3]:
            incar["ISYM"] = 0
            correction = "set ISYM to 0"

        # If the error shows up on the first ionic step, it is likely that
        # our bond distances are too small. A simple fix is to scale our
        # starting lattice. We also save the original structure to a file
        # name POSCAR_original in case it's needed elsewhere.
        else:
            poscar_filename = os.path.join(directory, "POSCAR")
            structure = Structure.from_file(poscar_filename)
            structure.to("POSCAR", poscar_filename + "_original")
            structure.apply_strain(0.2)
            structure.to("POSCAR", poscar_filename)
            correction = "scaled the structure lattice by +20%"

        # Check the current ICHARG setting, where default is 0
        # If the ICHARG is less than 10, then we want to delete the CHGCAR
        # and WAVECAR to ensure the next run is a clean start.
        current_icharg = incar.get("ICHARG", 0)
        if current_icharg < 10:
            os.remove(os.path.join(directory, "CHGCAR"))
            os.remove(os.path.join(directory, "WAVECAR"))
            correction += " and deleted CHGCAR and WAVECAR"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        # now return the correction made for logging
        return correction



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.database.base_data_types import Calculation as CalculationTable


class Calculation(filters.FilterSet):
    class Meta:
        model = CalculationTable
        fields = dict(
            directory=["exact"],
            prefect_flow_run_id=["exact"],
            created_at=["range"],
            updated_at=["range"],
        )



# -*- coding: utf-8 -*-

from numpy.random import randint, choice

from tqdm import tqdm

from simmate.toolkit.symmetry.wyckoff import (
    loadAsymmetricUnitData,
    loadWyckoffData,
    findValidWyckoffCombos,
)
from simmate.toolkit.creators.vector import UniformlyDistributedVectors


class RandomWySites:
    def __init__(
        self,
        composition,  # must be a pymatgen Composition object
        spacegroup_include=range(1, 231),
        spacegroup_exclude=[],
        coords_generation_method=UniformlyDistributedVectors,
        coords_gen_options=dict(),  # asymmetric_unit_boundries is include in extra_conditions automatically
    ):

        # make a list of spacegroups that we are allowed to choose from
        self.spacegroup_options = [
            sg for sg in spacegroup_include if sg not in spacegroup_exclude
        ]

        # save the composition for reference
        self.composition = composition

        # the stoichiometry of the composition will be used repeatedly below
        # it's easier to save that here for reference too
        # convert the composition to a stoichiometry list
        # for example, Mg4Si4O12 composition will have stoich=[4,4,12]
        stoichiometry = [int(composition[element]) for element in composition]
        self.stoichiometry = stoichiometry

        # Find valid wyckoff combinatons for all spacegroups and stoichiometry
        # I want the combinations and info to be stored as a dictionary of {spacegroup:combos/info} entries
        self.wy_groupinfo = {}
        self.wy_groupcombos = {}
        # some spacegroups will be incompatible with the given composition -- keep a list of these
        self.spacegroups_invalid = []
        #!!! NOTE: This code takes a long time to run in some cases, so I want to do it up front
        # because this takes time, I print a warning and then use tqdm to track progress
        message = "Generating Wyckoff combinations for every spacegroup: "
        for spacegroup in tqdm(self.spacegroup_options, desc=message):
            sg_combo_data = findValidWyckoffCombos(stoichiometry, spacegroup)
            # If the spacegroup + stoich combination has no valid combinations, the generator will not work
            if not sg_combo_data["ValidCombinations"]:
                self.spacegroups_invalid.append(spacegroup)
                # no need to save these results so skip to the next spacegroup
                continue
            # splitting up this dictionary becuase we reference each a lot
            # and append the results to respective dictionaries
            self.wy_groupinfo.update({spacegroup: sg_combo_data["WyckoffGroups"]})
            self.wy_groupcombos.update({spacegroup: sg_combo_data["ValidCombinations"]})
        # we need to take the spacegroups that are invalid and update the spacegroup_options with it.
        # we still save the self.spacegroups_invalid for the user to see #!!! consider removing self.spacegroups_invalid though
        self.spacegroup_options = [
            sg for sg in self.spacegroup_options if sg not in self.spacegroups_invalid
        ]

        # make all generators for each spacegroup
        # the generators are different for each spacegroup because each has a different asym unit
        #!!! NOTE: options set in coords_gen_options will apply to all spacegroups
        # I want the generators to be stored as a dictionary of {spacegroup:generator} entries
        self.coords_generators = {}
        for spacegroup in self.spacegroup_options:
            # grab a copy of the user inputs for reference
            # we will manipulate final_options for each spacegroup for use in the generator
            final_options = coords_gen_options.copy()
            # If the spacegroup was found to have no valid wyckoff combinations, there is no need for a generator
            if spacegroup in self.spacegroups_invalid:
                # skip to next spacegroup
                continue
            # establish the coords generator with asymmetric_unit_boundries
            # this process is a little more complicated because we require that the asymmetric_unit_boundries be included in extra_conditions
            asym_bounds = asymmetric_unit_boundries(spacegroup)
            # if any extra_conditions were provided by the user, we want to combine them with the asym conditions
            if "extra_conditions" in coords_gen_options.keys():
                all_bounds = coords_gen_options["extra_conditions"] + asym_bounds
                final_options.update({"extra_conditions": all_bounds})
            # otherwise, the asym_bounds are the only boundry conditions for the coordinates
            else:
                final_options.update({"extra_conditions": asym_bounds})
            # now that we have the coords_gen_options updated, we can make the generator
            coords_generator = coords_generation_method(**final_options)
            # add this generator to the results
            self.coords_generators.update({spacegroup: coords_generator})

        # below, I'll need to repeatedly reference wy_data
        #!!! I want to find a better way of loading wy_data - maybe load only data for spacegroup_options?
        self.wy_data = loadWyckoffData().values

    def new_sites(self, spacegroup=None):

        # This section of code is checking the spacegroup input
        # if a spacegroup is not specified, grab a random one from our options
        if not spacegroup:
            # randomly select a symmetry system
            spacegroup = choice(self.spacegroup_options)
        #!!! do I need to check this? This may slow the algorithm, but ensures a user doesn't break the function
        # if a spacegroup is specified, we still need to make sure it's valid
        else:
            check = spacegroup not in self.spacegroups_invalid
            # if it is invalid, print a message and quit the function
            if not check:
                # print('This spacegroup is invalid for the given composition') #!!! consider adding a warning
                return False

        # we need to grab the generators, wyckoff combinations, wyckoff group information for selected spacegroup
        coords_generator = self.coords_generators[spacegroup]
        wy_groupcombos = self.wy_groupcombos[spacegroup]
        wy_groupinfo = self.wy_groupinfo[spacegroup]

        # randomly grab a wy_group combination
        # NOTE: this does not have equal probability of grabbing all wyckoff site combinations because some wy_groups have more wy_site posibilities than others
        # choice(sg_combos['ValidCombinations']) doesn't work here, so we grab a random index instead
        random_index = randint(0, len(wy_groupcombos))
        wy_group_combo = wy_groupcombos[random_index]

        # make an empty list to store all the fractional coordinates
        coords_list = []  # regular list.append is faster than numpy.append
        # as well as an empty list for the species
        species_list = []

        # now we need to iterate through each element in stoich, see which wy_group(s) it was assigned to
        for i, wy_groups in enumerate(wy_group_combo):
            # wy_groups represents the combination of wy_group's for a single element
            # len(stoich) == len(wy_group_combo) because indexes are matched up (i.e. stoich[1] was assigned wy_group_combo[1])
            # so we can use the index of wy_groups to grab the element
            element = self.composition.elements[i]
            # wy_groups can be a list of wy_group entries
            for wy_group in wy_groups:
                # now we need to generate the coordinates for this wy_site
                # for that, we need three things:
                # (1) a random wy_site from the assigned wy_group
                # (2) the base wy_site coords corresponding to that wy_site (i.e. (x,y,z) or (0,y,0))
                # (3) random coords that exist inside the asymmetric unit, used to eval() the coords with
                # (4) ensure that the result coordinates have not been used already
                # wy_group label is a key to the wy_groupinfo dictionary
                # the wy_sites output is a list of indexes (as numpy array) that correspond to indicies in the loadWyckoffData() output
                wy_sites = wy_groupinfo[wy_group]
                check = False
                while not check:
                    # randomly choose one of these wy_site indexes
                    # I have this inside the while-loop because if a special wy_site (i.e. (0,0,0)) is picked when it's already in coords_used, we will get stuck in the while loop
                    wy_site_i = choice(wy_sites)
                    # use the choice to grab the wy_site
                    wy_site = self.wy_data[wy_site_i]
                    # grab the first entry (as all others are symmetrically equivalent) for coord template
                    wy_coords = wy_site[5]  # index 5 is 'Coordinates'
                    # generate random x,y,z values that are inside the asymmetric unit
                    x, y, z = coords_generator.new_vector()
                    # place these x,y,z values into the coords where necessary
                    # not sure why I need to use None,dict(x=x,y=y,z=z) inside eval. Eval() can't use the locals for some reason.
                    final_coords = eval(wy_coords, None, dict(x=x, y=y, z=z))
                    # see if the result is already a coordinate used
                    check = final_coords not in coords_list
                # append the coords and species to our output lists
                coords_list.append(final_coords)
                species_list.append(element)

        # maybe make these into pymatgen Site objects? I don't see any advantage to Site object,
        # because we will later need PeriodicSite objects and there's no direct conversion method
        return species_list, coords_list


##############################################################################

# Grab the boundry conditions for the asymmetric unit of a spacegroup's unitcell
def asymmetric_unit_boundries(spacegroup, asym_data=loadAsymmetricUnitData()):

    # spacegroup = spacegroup that we are interested in
    # asym_data = result of loadAsmmetricUnitData() - this is in the header for speed reasons

    # grab the asymmetric unit conditons from the asym_data
    # the array indicies (0-229) correspond to all spacegroups (1-230) respectively.
    # So if I want data for spacegroup 230, I need index 229 (i.e. spacegroup-1)
    asym_bounds = asym_data[spacegroup - 1]

    # convert the string of asym_data into a list of boundry conditions that we can test
    # so here we change things to something python can read and iterate through
    asym_bounds = asym_bounds.replace("≤", "<=")  # convert symbols
    # Is there a way I can do these next three lines in a single command...?
    # or maybe only run them for the spacegroups that need them?
    asym_bounds = asym_bounds.replace("2y", "(2*y)")  # convert
    asym_bounds = asym_bounds.replace("2x", "(2*x)")  # convert
    asym_bounds = asym_bounds.replace("-4x", "(-4*y)")  # convert
    asym_bounds = asym_bounds.split(";")  # convert str into a list

    return asym_bounds



# -*- coding: utf-8 -*-

##############################################################################

from numpy.random import choice

from pymatgen.core.lattice import Lattice

from simmate.toolkit.creators.vector import (
    UniformlyDistributedVectors,
    NormallyDistributedVectors,
)

##############################################################################

# Generates a random lattice without symmetry
# this is the same as RandomSymLattice when spacegroup is assumed to be 1
class RandomLattice:
    def __init__(
        self,
        vector_generation_method=NormallyDistributedVectors,
        vector_gen_options=dict(),
        angle_generation_method=UniformlyDistributedVectors,
        angle_gen_options=dict(min_value=60, max_value=120),
    ):

        # establish the vector generation method and use the specified options
        self.vector_generator = vector_generation_method(**vector_gen_options)

        # establish the vector generation method and use the specified options
        self.angle_generator = angle_generation_method(**angle_gen_options)

    def new_lattice(self):

        # generate an (a,b,c) vector to pull lattice vectors from
        a, b, c = self.vector_generator.new_vector()

        # generate an (alpha,beta,gamme) vector to pull lattice angles from
        alpha, beta, gamma = self.angle_generator.new_vector()

        # take the inputs of vectors=(a,b,c) and angles=(alpha,beta,gamma) and create a pymatgen lattice object
        lattice = Lattice.from_parameters(a, b, c, alpha, beta, gamma)

        return lattice


##############################################################################

# Generates a random lattice with symmetry
class RandomSymLattice:
    def __init__(
        self,
        spacegroup_include=range(1, 231),
        spacegroup_exclude=[],
        vector_generation_method=NormallyDistributedVectors,
        vector_gen_options=dict(),
        angle_generation_method=UniformlyDistributedVectors,
        angle_gen_options=dict(min_value=60, max_value=120),
    ):

        # sg_include = list of spacegroups that we are interested in. Default is all 230 spacegroups
        # sg_exclude = list of spacegroups that we should explicitly ignore

        # establish the vector generation method and use the specified options
        self.vector_generator = vector_generation_method(**vector_gen_options)

        # establish the vector generation method and use the specified options
        self.angle_generator = angle_generation_method(**angle_gen_options)

        # make a list of spacegroups that we are allowed to choose from
        self.spacegroup_options = [
            sg for sg in spacegroup_include if sg not in spacegroup_exclude
        ]

    def new_lattice(self, spacegroup=None):

        # if a spacegroup is not specified, grab a random one from our options
        # no check is done to see if the spacegroup specified is compatible with the vector_generator built
        if not spacegroup:
            # randomly select a symmetry system
            spacegroup = choice(self.spacegroup_options)

        # generate an (a,b,c) vector to pull lattice vectors from
        a, b, c = self.vector_generator.new_vector()

        # generate an (alpha,beta,gamme) vector to pull lattice angles from
        alpha, beta, gamma = self.angle_generator.new_vector()

        # Now generate the lattice using the spacegroup indicidated
        if spacegroup <= 2:  # triclinic
            # SPECIAL CASE b/c pymatgen does not have this function
            # but because (a != b != c) and (alpha != beta != gamma)
            # take the inputs of vectors=(a,b,c) and angles=(alpha,beta,gamma) and create a pymatgen lattice object
            lattice = Lattice.from_parameters(a, b, c, alpha, beta, gamma)
        elif spacegroup <= 15:  # monoclinic
            lattice = Lattice.monoclinic(a, b, c, beta)
        elif spacegroup <= 74:  # orthorhombic
            lattice = Lattice.orthorhombic(a, b, c)
        elif spacegroup <= 142:  # tetragonal
            lattice = Lattice.tetragonal(a, c)
        elif spacegroup <= 167:  # trigonal
            lattice = Lattice.hexagonal(a, c)
            # Note: I have all lattices in range(143,168) to be hexagonal
            # the spacegroups 146,148,155,160,161,166,167 can optionally be rhombohedral though
            # lattice = Lattice.rhombohedral(a,alpha)
        elif spacegroup <= 194:  # hexagonal
            lattice = Lattice.hexagonal(a, c)
        elif spacegroup <= 230:
            lattice = Lattice.cubic(a)

        # we need to know which spacegroup was used in addition to the result lattice
        # maybe I can add an option to 'returnlatticeonly=True' if I don't want this as a dict
        return lattice


##############################################################################


class RSLFixedVolume(RandomSymLattice):

    # identical to RandomSymLattice, except the lattice is scaled to a fixed volume at the end

    #!!! currently, this takes roughly 3x as long as the RandomSymLattice to run (180us vs 55.3us averages for new_lattice)
    #!!! this time increase is primarily in the .scale() function (when I comment out the check loop - I get a 154us average)

    #!!! this is fixed volume of the conventional cells! So some spacegroups may run into issues with this.
    #!!! I need a 'smart' FixedVolume that scales based on...
    # composition (total atoms and their type)
    # spacegroup (conventional vs primitive size)

    def __init__(self, volume, **kwargs):
        # run the same parent init
        super().__init__(**kwargs)

        # now do the extra code
        # which is just save the volume here
        self.volume = volume

    def new_lattice(self, spacegroup=None):
        # run the same parent new_lattice, which returns a lattice
        lattice = super().new_lattice(spacegroup)

        # now scale the lattice to the specified volume
        lattice = lattice.scale(self.volume)

        # in scaling, we might have broken the conditions of min/max_vectors
        # I check this and if it fails, I scrap the lattice and try making a new one via recursion
        check = True  # assume valid until proven otherwise
        for vector in [lattice.a, lattice.b, lattice.c]:
            if vector < self.vector_generator.min_value:
                check = False
                break
            elif vector > self.vector_generator.max_value:
                check = False
                break
        if not check:
            # this is a recursion call that will be hit repeatedly until a valid lattice is found
            #!!! this will throw a recursion error if the volume and min/max_vectors are unreasonable
            lattice = self.new_lattice(spacegroup)

        return lattice


##############################################################################

from simmate.toolkit.symmetry.wyckoff import loadWyckoffData

#!!! EASY SPEED/MEMORY IMPROVEMENTS CAN BE MADE HERE (ON INIT)
class RSLSmartVolume:
    def __init__(
        self,
        composition,
        volume=None,
        spacegroup_include=range(1, 231),
        spacegroup_exclude=[],
        vector_generation_method=NormallyDistributedVectors,
        vector_gen_options=dict(),
        angle_generation_method=UniformlyDistributedVectors,
        angle_gen_options=dict(min_value=60, max_value=120),
    ):

        # make a list of spacegroups that we are allowed to choose from
        self.spacegroup_options = [
            sg for sg in spacegroup_include if sg not in spacegroup_exclude
        ]

        # if the user didn't specify a volume, we should predict our own
        if not volume:
            # grab the base predicted volume which is based on the composition
            # note this represents the volume of the primitive cell!!! Conventional cells may be larger
            #!!! this assumes user wants ionic radius estimation
            volume = composition.volume_estimate()

        # find the respective volume for each spacegroup's conventional unit cell
        # to do this, we need to know the ratio between MultiplicityPrimitive and MultiplicityConventiional for a spacegroup
        #!!! currently this is stored in wyckoffdata.csv, but I should consider putting it in a separate file for speed
        data = loadWyckoffData()
        self.volumes = {}  # keys are spacegroup
        self.vector_generators = {}  # keys are spacegroup
        for spacegroup in self.spacegroup_options:
            # grab the portion of the table for spacegroup and make it into a numpy array
            sg_data = data.query("SpaceGroup == @spacegroup").values
            # final the ratio between the primitive and conventional unitcells for this structure
            # this is the ratio between the 2nd and 3rd columns (indicies 1 and 2), which are MultiplicityConventional & MultiplicityPrimitive
            # I grab the first row, which is the first wyckoff site -- it makes no difference which one I grab
            ratio = sg_data[0][1] / sg_data[0][2]
            # the volume of the conventional cell for this spacegroup is the primitive times this ratio
            self.volumes.update({spacegroup: volume * ratio})

            # now we want to make the vector_generator for this spacegroup
            #!!! I'm not sure users can pass custom parameters this deep and if they will work. I need to test for this
            #!!! I assume NormallyDistributedVectors is used here... will give errors if that's not the case
            vgen = vector_generation_method(
                min_value=min(
                    composition.radii_estimate()
                ),  # smallest radius of any ion #!!! move outside for loop for speed! #!!! this assumes user wants ionic radius estimation
                center=(ratio * volume) ** (1 / 3),  # shoot for cubic vectors
                max_value=(ratio * volume)
                ** 0.8,  # don't allow for vectors too be crazy different #!!! need to test if 0.8 is a good value here
                standdev=((ratio * volume) ** (1 / 3))
                * 0.15,  # larger volumes have more flexibility in their vector lengths
            )  #!!! **vector_gen_options not used... need to fix
            self.vector_generators.update({spacegroup: vgen})

        # establish the vector generation method and use the specified options
        # Unlike the vector_generators, we can use the same generator here for all spacegroups
        self.angle_generator = angle_generation_method(**angle_gen_options)

    def new_lattice(self, spacegroup=None):

        # if a spacegroup is not specified, grab a random one from our options
        # no check is done to see if the spacegroup specified is compatible with the vector_generator built
        if not spacegroup:
            # randomly select a symmetry system
            spacegroup = choice(self.spacegroup_options)

        # For our target spacegroup, grab the target volume and vector_generator
        volume = self.volumes[spacegroup]
        vector_generator = self.vector_generators[spacegroup]

        # run the same parent new_lattice, which returns a lattice
        # if a spacegroup is not specified, grab a random one from our options
        # no check is done to see if the spacegroup specified is compatible with the vector_generator built
        if not spacegroup:
            # randomly select a symmetry system
            spacegroup = choice(self.spacegroup_options)

        # generate an (a,b,c) vector to pull lattice vectors from
        a, b, c = vector_generator.new_vector()

        # generate an (alpha,beta,gamma) vector to pull lattice angles from
        alpha, beta, gamma = self.angle_generator.new_vector()

        # Now generate the lattice using the spacegroup indicidated
        if spacegroup <= 2:  # triclinic
            # SPECIAL CASE b/c pymatgen does not have this function
            # but because (a != b != c) and (alpha != beta != gamma)
            # take the inputs of vectors=(a,b,c) and angles=(alpha,beta,gamma) and create a pymatgen lattice object
            lattice = Lattice.from_parameters(a, b, c, alpha, beta, gamma)
        elif spacegroup <= 15:  # monoclinic
            lattice = Lattice.monoclinic(a, b, c, beta)
        elif spacegroup <= 74:  # orthorhombic
            lattice = Lattice.orthorhombic(a, b, c)
        elif spacegroup <= 142:  # tetragonal
            lattice = Lattice.tetragonal(a, c)
        elif spacegroup <= 167:  # trigonal
            lattice = Lattice.hexagonal(a, c)
            # Note: I have all lattices in range(143,168) to be hexagonal
            # the spacegroups 146,148,155,160,161,166,167 can optionally be rhombohedral though
            # lattice = Lattice.rhombohedral(a,alpha)
        elif spacegroup <= 194:  # hexagonal
            lattice = Lattice.hexagonal(a, c)
        elif spacegroup <= 230:
            lattice = Lattice.cubic(a)

        #!!! This section makes the code take more than 3x longer to run. Should
        #!!! I allow a more flexible volume? More flexible volume could also slow structure creation (bc smaller volumes are allowed)
        # now scale the lattice to the specified volume
        lattice = lattice.scale(volume)
        # in scaling, we might have broken the conditions of min/max_vectors
        # I check this and if it fails, I scrap the lattice and try making a new one via recursion
        check = True  # assume valid until proven otherwise
        for vector in [lattice.a, lattice.b, lattice.c]:
            if vector < vector_generator.min_value:
                check = False
                break
            elif vector > vector_generator.max_value:
                check = False
                break
        if not check:
            # this is a recursion call that will be hit repeatedly until a valid lattice is found
            #!!! this will throw a recursion error if the volume and min/max_vectors are unreasonable
            #!!! I need to rewrite this function where it's a while loop instead of a recursion call
            lattice = self.new_lattice(spacegroup)

        # we need to know which spacegroup was used in addition to the result lattice
        # maybe I can add an option to 'returnlatticeonly=True' if I don't want this as a dict
        return lattice



# -*- coding: utf-8 -*-

import numpy as np

from matminer.featurizers.site import CrystalNNFingerprint as cnnf

#!!! is it faster to import these all at once?
from matminer.featurizers.structure import SiteStatsFingerprint as ssf
from matminer.featurizers.structure import RadialDistributionFunction as rdf
from matminer.featurizers.structure import PartialRadialDistributionFunction as prdf


class CrystalNNFingerprint:

    # requires matminer code
    # see https://materialsproject.org/docs/structuresimilarity

    def __init__(
        self,
        on_fail,
        max_fails,
        cnn_options={},
        stat_options=["mean", "std_dev", "minimum", "maximum"],
        initial_structures=[],
        parallel=False,
    ):

        self.on_fail = on_fail

        # we need the cumulative max_fails list
        # for example, the max_fails=[20,40,10] needs to be converted to [20,60,70]
        # if infinite attempts are allowed, be sure to use float('inf') as the value
        # This list is more efficient to work with in the point_next_step()
        #!!! should I make this numpy for speed below?
        self.max_fails_cum = [sum(max_fails[: x + 1]) for x in range(len(max_fails))]

        # make the matminer featurizer object
        # if it isnt set, we will use the default, which is set to what the Materials project uses
        if cnn_options:
            sitefingerprint_method = cnnf(**cnn_options)
        else:
            # see https://materialsproject.org/docs/structuresimilarity
            sitefingerprint_method = cnnf.from_preset(
                "ops", distance_cutoffs=None, x_diff_weight=0
            )
        # now that we made the sitefingerprint_method, we can input it into the structurefingerprint_method which finishes up the featurizer
        self.featurizer = ssf(sitefingerprint_method, stats=stat_options)

        # generate the fingerprint for each of the initial input structures
        # note this uses the parallel method for the matminer featurizer
        # if there are less the 15 structures, parallel method is actually slower than serial
        # so we check that first and decide which is quickest
        # sometimes the script will stall/fail when trying parallel (like when in Spyder) so the option to turn of parallel is also there
        if len(initial_structures) <= 15 and not parallel:
            # do this serially and make into a numpy array for speed
            self.fingerprint_database = np.array(
                [
                    self.featurizer.featurize(structure)
                    for structure in initial_structures
                ]
            )
        else:
            # use matminer's parallel functionality and make into a numpy array for speed
            self.fingerprint_database = self.featurizer.featurize_many(
                initial_structures, pbar=False
            )

    def check_structure(self, structure, tolerance=1e-4):

        # make the fingerprint for this structure and make into a numpy array for speed
        fingerprint1 = np.array(self.featurizer.featurize(structure))

        # we now want to get the distance of this fingerprint relative to all others
        # if the distance is within the specified tolerance, then the structures are too similar - we return false
        # if we make it through all structures and no distance is below the tolerance, we are good to go!
        for fingerprint2 in self.fingerprint_database:
            distance = np.linalg.norm(fingerprint1 - fingerprint2)
            if distance < tolerance:
                # we can end the whole function as soon as one structure is deemed too similar
                return False
            else:
                continue

        # if we finished the loop above, we can add this structure to the database and return a success
        # this method I use to append is pretty slow for numpy. Is there a better way to do this?
        if self.fingerprint_database.size == 0:
            self.fingerprint_database = np.array([fingerprint1])
        else:
            self.fingerprint_database = np.append(
                self.fingerprint_database, [fingerprint1], axis=0
            )

        return True



# -*- coding: utf-8 -*-

import os
import pytest

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import MeshSymmetry


def test_mesh_symmetry(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="mesh_symmetry",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    kpoints_filename = os.path.join(tmpdir, "KPOINTS")
    vasprun_filename = os.path.join(tmpdir, "vasprun.xml")

    # init class with default settings
    error_handler = MeshSymmetry()

    # Confirm an error IS NOT found when file doesn't exist
    error_handler.filename_to_check = "vasp.does_not_exist"
    assert error_handler.check(tmpdir) == False

    # Confirm an error IS NOT found when vasprun.xml is converged
    error_handler.filename_to_check = "vasp.out"
    assert error_handler.check(tmpdir) == False

    # Confirm an error IS found when vasprun.xml is not converged
    os.remove(vasprun_filename)
    assert error_handler.check(tmpdir) == True

    # Make attempt at fixing the error
    with pytest.raises(NotImplementedError):
        error_handler.correct(tmpdir)

    # Confirm an error IS NOT found when KPOINTS file doesn't exist and KSPACING
    # was set instead.
    os.remove(kpoints_filename)
    incar = Incar.from_file(incar_filename)
    incar["KSPACING"] = 0.3
    incar.to_file(incar_filename)
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

import pytest

from simmate.website.test_app.models import TestCalculation


@pytest.mark.django_db
def test_calculation_table():

    # test writing columns
    TestCalculation.show_columns()

    # test writing to database
    calc_db = TestCalculation.from_prefect_id(prefect_flow_run_id="example-id-123")
    calc_db.save()

    # try grabbing the calculation again and make sure it loaded from the
    # database rather than creating a new entry
    calc_db2 = TestCalculation.from_prefect_id(prefect_flow_run_id="example-id-123")
    assert calc_db.id == calc_db2.id

    # grab prefect url for this id
    assert (
        calc_db.prefect_cloud_link
        == "https://cloud.prefect.io/simmate/flow-run/example-id-123"
    )


@pytest.mark.django_db
def test_calculation_archives():

    calc_db = TestCalculation.from_prefect_id(prefect_flow_run_id="example-id-123")
    calc_db.save()
    calc_db2 = TestCalculation.from_prefect_id(prefect_flow_run_id="example-id-321")
    calc_db2.save()

    TestCalculation.objects.to_archive()

    TestCalculation.load_archive(
        confirm_override=True,
        delete_on_completion=True,
    )



# -*- coding: utf-8 -*-

"""
Example use:
``` python
from simmate.toolkit import Structure
from simmate.workflows.static_energy import mit_workflow
from simmate.workflows.customized import vasp_workflow

s = Structure(
    lattice=[
        [3.485437, 0.0, 2.012318],
        [1.161812, 3.286101, 2.012318],
        [0.0, 0.0, 4.024635],
    ],
    species=["Na", "Cl"],
    coords=[[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]],
)

vasp_workflow.run(
    workflow_base=mit_workflow,
    input_parameters={"structure": s},
    updated_settings={
        "incar": {"NPAR": 1, "ENCUT": 600},
    },
)
```
"""

from simmate.workflow_engine import Workflow
from simmate.workflow_engine.common_tasks import (
    load_input_and_register,
    run_customized_s3task,
    save_result,
)
from simmate.calculators.vasp.database.customized import (
    CustomizedVASPCalculation,
)


class Customized__Vasp__UserConfig(Workflow):
    """
    "VASP calculation with updated INCAR/POTCAR settings
    """

    database_table = CustomizedVASPCalculation

    @staticmethod
    def run_config(
        workflow_base: str,
        input_parameters: dict,
        updated_settings: dict,
    ):
        parameters_cleaned = load_input_and_register(
            workflow_base=workflow_base,
            input_parameters=input_parameters,
            updated_settings=updated_settings,
        )

        result = run_customized_s3task(**parameters_cleaned)

        save_result(result)



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure


class OqmdStructure(Structure):
    """
    Crystal structures from the [OQMD](http://oqmd.org/) database.

    Currently, this table only stores strucure and thermodynamic information,
    but OQDMD has much more data available via their
    [REST API](http://oqmd.org/static/docs/restful.html) and website.
    """

    class Meta:
        app_label = "third_parties"

    base_info = ["id", "structure_string", "formation_energy"]
    source = "The Open Quantum Materials Database"
    source_doi = "https://doi.org/10.1007/s11837-013-0755-4"
    remote_archive_link = "https://archives.simmate.org/OqmdStructure-2022-02-22.zip"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "oqmd-12345")
    """

    # OQMD did not provide energy so the Thermodynamics mix-in can't be used
    formation_energy = table_column.FloatField(blank=True, null=True)
    """
    The formation energy of the structure as provided by the OQMD.
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the OQMD website.
        """
        # Links to the OQMD dashboard for this structure. An example is...
        #   http://oqmd.org/materials/entry/10435
        id_number = self.id.split("-")[-1]  # this removes the "oqmd-" from the id
        return f"http://oqmd.org/materials/entry/{id_number}"



# -*- coding: utf-8 -*-

from .quality_00 import Quality00Relaxation
from .quality_01 import Quality01Relaxation
from .quality_02 import Quality02Relaxation
from .quality_03 import Quality03Relaxation
from .quality_04 import Quality04Relaxation

from .mit import MITRelaxation

from .materials_project import MatprojRelaxation
from .materials_project_hse import MatprojHSERelaxation
from .materials_project_metal import MatprojMetalRelaxation

from .materials_virtual_lab_endpoint import MatVirtualLabCINEBEndpointRelaxation
from .materials_virtual_lab_grain_boundary import MatVirtualLabGrainBoundaryRelaxation
from .materials_virtual_lab_slab import MatVirtualLabSlabRelaxation



# -*- coding: utf-8 -*-

import numpy

from simmate.toolkit import Structure as ToolkitStructure
from simmate.database.base_data_types import DatabaseTable, table_column


class Forces(DatabaseTable):
    class Meta:
        abstract = True

    base_info = [
        "site_forces",
        "lattice_stress",
    ]

    site_forces = table_column.JSONField(blank=True, null=True)
    """
    A list of forces for each atomic site (eV/AA). So this is a list like...
    ```
    [site1, site2, site3, ...] 
    ```
    ... where ...
    ```
    site1=[force_x, force_y, force_z]
    ```
    """

    lattice_stress = table_column.JSONField(blank=True, null=True)
    """
    The is 3x3 matrix that represents stress on the structure lattice
    """

    site_force_norm_max = table_column.FloatField(blank=True, null=True)
    """
    Reports the vector norm for the site with the highest forces on it
    """

    site_forces_norm = table_column.FloatField(blank=True, null=True)
    """
    Takes the site forces and reports the vector norm for it.

    See numpy.linalg.norm for how this is calculated.
    """

    site_forces_norm_per_atom = table_column.FloatField(blank=True, null=True)
    """
    site_forces_norm divided by nsites
    """

    lattice_stress_norm = table_column.FloatField(blank=True, null=True)
    """
    Takes the site forces and reports the vector norm for it.

    See numpy.linalg.norm for how this is calculated.
    """

    lattice_stress_norm_per_atom = table_column.FloatField(blank=True, null=True)
    """
    lattice_stress_norm divided by nsites
    """

    @classmethod
    def _from_toolkit(
        cls,
        structure: ToolkitStructure,
        site_forces=None,
        lattice_stress=None,
        as_dict=False,
    ):
        """
        Given site forces and lattice stress, this function builds the rest of
        the required fields for this class as a dictionary.
        """
        # TODO: in the future, this should accept an IonicStep toolkit object
        # or maybe Structure + Forces toolkit objects.

        site_data = (
            dict(
                site_forces=site_forces,
                site_force_norm_max=max([numpy.linalg.norm(f) for f in site_forces]),
                site_forces_norm=numpy.linalg.norm(site_forces),
                site_forces_norm_per_atom=numpy.linalg.norm(site_forces)
                / structure.num_sites,
            )
            if site_forces
            else {}
        )

        lattice_data = (
            dict(
                lattice_stress=lattice_stress,
                lattice_stress_norm=numpy.linalg.norm(lattice_stress),
                lattice_stress_norm_per_atom=numpy.linalg.norm(lattice_stress)
                / structure.num_sites,
            )
            if lattice_stress
            else {}
        )

        all_data = dict(**site_data, **lattice_data)

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return all_data if as_dict else cls(**all_data)



# -*- coding: utf-8 -*-

import os
import yaml
from typing import Any

from prefect import task
from prefect.context import FlowRunContext

from simmate.utilities import get_directory, copy_directory
from simmate.workflow_engine import Workflow

# OPTIMIZE: consider splitting this task into load_structure, load_directory,
# and register_calc so that our flow_visualize looks cleaner

# OPTIMIZE: Customized workflows cause a lot of special handling in this task
# so it may be worth isolating these into a separate task.


@task
def load_input_and_register(
    register_run: bool = True,
    **parameters: Any,
) -> dict:
    """
    How the input was submitted as a parameter depends on if we are submitting
    to Prefect Cloud, running the flow locally, or even continuing from a
    previous calculation.  Here, we use a task to convert the input to a toolkit
    object and (if requested) provide the directory as well.

    For example, a common input parameter for workflows is "structure", which
    can be provided a number of ways:
        - a filename
        - a json string
        - a dictionary pointing to a database entry
        - a toolkit Structure object
        - etc...
    Even though all of these inputs are accepted, `workflow.run` always expects
    python objects, so this utility converts the input to a toolkit Structure
    object.

    `register_run` allows us to skip the database step if the database_table
    isn't properly set yet. This input is a temporary fix for the
    diffusion/from-images workflow.

    `copy_previous_directory` is only used when we are pulling a structure from a
    previous calculation. If copy_previous_directory=True, then the directory
    parameter is ignored.

    `**parameters` includes all parameters and anything extra that you want saved
    to simmate_metadata.yaml
    """

    # ---------------------------------------------------------------------

    # Grab the workflow object as we need to reference some of its attributes.
    # In addition, we will also use the flow run id for registration.
    run_context = FlowRunContext.get()
    workflow_name = run_context.flow.name
    prefect_flow_run_id = str(run_context.flow_run.id)
    workflow = run_context.flow.simmate_workflow

    # ---------------------------------------------------------------------

    # STEP 1: clean parameters

    parameters_cleaned = Workflow._deserialize_parameters(**parameters)

    # ---------------------------------------------------------------------

    # STEP 1b: Determine the "primary" input to use for setting the
    # source (and previous directory)
    # OPTIMIZE: Is there a better way to do this?

    # Currently I just set a priority of possible parameters that can be
    # the primary input. I go through each one at a time until I find one
    # that was provided -- then I exit with that parameter's value.
    primary_input = None
    for primary_input_key in ["structure", "migration_hop", "supercell_start"]:
        primary_input = parameters.get(primary_input_key, None)
        primary_input_cleaned = parameters_cleaned.get(primary_input_key, None)
        if primary_input:
            break

    # ---------------------------------------------------------------------

    # STEP 2: Load the directory (and copy over from an old directory if necessary)

    # Start by creating a new directory or grabbing the one given. We create
    # this directory immediately (rather than just passing the name to the
    # S3Task). We do this because NestedWorkflows often use a parent directory
    # to organize results.
    directory = parameters.get("directory", None)
    directory_cleaned = get_directory(directory)

    # if the user requested, we grab the previous directory as well
    copy_previous_directory = parameters.get("copy_previous_directory", None)
    if copy_previous_directory:

        if not primary_input:
            raise Exception(
                "No primary input detected, which is required for copying "
                "past directories. This is an experimental feature so "
                "please contact our team for more help."
            )

        # catch incorrect use of this function
        if not primary_input_cleaned.is_from_past_calc:
            raise Exception(
                "There isn't a previous directory available! Your source "
                "structure must point to a past calculation to use this feature."
            )

        # the past directory should be stored on the input object
        previous_directory = primary_input_cleaned.database_object.directory

        # Copy over all files except simmate one (we have no need for the
        # summaries or error archives)
        copy_directory(
            directory_old=previous_directory,
            directory_new=directory_cleaned,
            ignore_simmate_files=True,
        )

    # SPECIAL CASE for customized flows
    if "workflow_base" not in parameters_cleaned:
        parameters_cleaned["directory"] = directory_cleaned
    else:
        parameters_cleaned["input_parameters"]["directory"] = directory_cleaned

    # ---------------------------------------------------------------------

    # STEP 3: Load the source of the input object

    source = parameters.get("source", None)

    # If we were given a input from a previous calculation, the source should
    # point directory to that same input. Otherwise we are incorrectly trying
    # to change what the source is.
    # "primary_input and" is added to the start to ensure cleaned input exists
    # and therefore prevent an error/bug.
    if source and primary_input and primary_input_cleaned.is_from_past_calc:
        # note primary_input here is a dictionary
        # assert
        if not source == primary_input:
            # only warning for now because this is experimental
            print(
                "\nWARNING: Your source does not match the source of your "
                "primary input. Sources are an experimental feature, so "
                "this will not affect your results. Still, please report "
                "this to our team to help with development. \n\n"
                f"SOURCE: {source} \n\n"
                f"PRIMARY_INPUT: {primary_input} \n\n"
            )
        source_cleaned = source
    # Check if we have a primary input loaded from a past calculation and
    # default to that as the source.
    elif primary_input and primary_input_cleaned.is_from_past_calc:
        source_cleaned = primary_input
    # Otherwise just use the source given
    elif source:
        source_cleaned = source
    else:
        source_cleaned = None

    # SPECIAL CASE for customized flows
    if "workflow_base" not in parameters_cleaned:
        parameters_cleaned["source"] = source_cleaned
    else:
        parameters_cleaned["input_parameters"]["source"] = source_cleaned

    # ---------------------------------------------------------------------

    # STEP 4: Register the calculation so the user can follow along in the UI
    # and also see which structures/runs have been submitted aready.

    if register_run:
        Workflow._register_calculation(**parameters_cleaned)

    # ---------------------------------------------------------------------

    # STEP 5: Write metadata file for user reference

    # convert back to json format. We convert back rather than use the original
    # to ensure the input data is all present. For example, we want to store
    # structure data instead of a filename in the metadata.
    # SPECIAL CASE: "if ..." used to catch customized workflows
    parameters_serialized = (
        Workflow._serialize_parameters(**parameters_cleaned)
        if "workflow_base" not in parameters_cleaned
        else parameters
    )

    # We want to write a file summarizing the inputs used for this
    # workflow run. This allows future users to reproduce the results if
    # desired -- and it also allows us to load old results into a database.
    input_summary = dict(
        workflow_name=workflow_name,
        # this ID is ingored as an input but needed for loading past data
        prefect_flow_run_id=prefect_flow_run_id,
        **parameters_serialized,
    )

    # now write the summary to file in the same directory as the calc.
    input_summary_filename = os.path.join(directory_cleaned, "simmate_metadata.yaml")
    with open(input_summary_filename, "w") as file:
        content = yaml.dump(input_summary)
        file.write(content)

    # ---------------------------------------------------------------------

    # Finally we just want to return the dictionary of cleaned parameters
    # to be used by the workflow
    return parameters_cleaned



# -*- coding: utf-8 -*-

from simmate.toolkit import Structure
from simmate.toolkit.diffusion import MigrationImages
from simmate.workflow_engine import task


@task
def get_migration_images_from_endpoints(supercell_start, supercell_end):
    """
    Simple wrapper for from_endpoints method that makes it a Prefect task.
    I assume parameters for now.
    """

    # Make sure we have toolkit objects, and if not, convert them
    supercell_start_cleaned = Structure.from_dynamic(supercell_start)
    supercell_end_cleaned = Structure.from_dynamic(supercell_end)

    # Then generate the images
    images = MigrationImages.from_endpoints(
        structure_start=supercell_start_cleaned,
        structure_end=supercell_end_cleaned,
        nimages=5,  # TODO: have from_endpoints figure out pathway length
    )

    return images



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.static_energy import (
    NEBEndpointStaticEnergy as NEBEndpointStaticEnergyTask,
)
from simmate.calculators.vasp.database.energy import (
    NEBEndpointStaticEnergy as NEBEndpointStaticEnergyResults,
)


class StaticEnergy__Vasp__NebEndpoint(Workflow):
    s3task = NEBEndpointStaticEnergyTask
    database_table = NEBEndpointStaticEnergyResults
    description_doc_short = (
        "uses Materials Project settings meant for defect supercell structures"
    )



# -*- coding: utf-8 -*-

import anyio
import time
import uuid
from functools import cached_property

from prefect.client import get_client
from prefect.exceptions import PrefectHTTPStatusError
from prefect.agent import OrionAgent
from prefect.settings import PREFECT_AGENT_QUERY_INTERVAL
from prefect.orion.schemas.filters import FlowRunFilter  # FlowFilter,

from simmate.utilities import async_to_sync

# This string is just something fancy to display in the console when a worker
# starts up.
# This uses "Small Slant" from https://patorjk.com/software/taag/
HEADER_ART = r"""
   _____                  __        _      __         __
  / __(_)_ _  __ _  ___ _/ /____   | | /| / /__  ____/ /_____ ____
 _\ \/ /  ' \/  ' \/ _ `/ __/ -_)  | |/ |/ / _ \/ __/  '_/ -_) __/
/___/_/_/_/_/_/_/_/\_,_/\__/\__/   |__/|__/\___/_/ /_/\_\\__/_/

"""


class Worker:
    """
    A worker is a process that checks the Prefect database for scheduled workflow
    runs and then submits them.

    Workers are only required when you are using a workflow's `run_cloud` method.

    This is effectively a combination of a Prefect Agent and Queue, where we
    configure default values for a queue and automatically perform the setup.
    """

    def __init__(
        self,
        queue_name: str = None,
        concurrency_limit: int = 1,
        nflows_max: int = None,
        timeout: float = None,
        close_on_empty_queue: bool = False,
    ):
        # This will be used when creating the queue
        self.concurrency_limit = concurrency_limit

        # if no name was given, we generate a random one
        if not queue_name:
            # generate a unique name for the worker, where we use uuid and take
            # the first 8 values (the rest is overkill for this application)
            unique_id = str(uuid.uuid4())[:8]
            queue_name = f"simmate-worker-{unique_id}"

        self.queue_name = queue_name

        # Grab the worker queue.
        # This is a cached property, so simply calling it and asserting that
        # it exists is sufficient to set everything up.
        assert self.queue_id

        # the maximum number of workitems to run before closing down
        # if no limit was set, we can go to infinity!
        self.nflows_max = nflows_max or float("inf")

        # Don't start a new workitem after this time. The worker will be shut down.
        # if no timeout was set, use infinity so we wait forever.
        self.timeout = timeout or float("inf")

        # whether to close if the queue is empty
        self.close_on_empty_queue = close_on_empty_queue

    @async_to_sync
    async def run(self):
        """
        Starts the worker and begins running scheduled workflows.
        """

        # !!! This method is really a fork of Prefect's `prefect agent start`
        # command, so a lot of this could be move to a prefect contribution.
        # https://github.com/PrefectHQ/prefect/blob/orion/src/prefect/cli/agent.py

        # print the header in the console to let the user know the worker started
        print(HEADER_ART)

        # establish starting point for the worker
        time_start = time.time()
        nflows_submitted = 0

        async with OrionAgent(work_queue_name=self.queue_name) as agent:

            print(f"Worker started! Looking for work from queue '{self.queue_name}'...")

            # Loop endlessly until one of the following happens...
            #   the timeout limit is hit
            #   the queue is empty
            #   the nitems limit is hit
            #   the user stops the process with ctrl+C  # BUG
            while True:

                # check for timeout before starting a new workitem and exit
                # if we've hit the limit.
                if (time.time() - time_start) > self.timeout:
                    print("The time-limit reached. Shutting down worker.")
                    break

                # check the number of jobs completed so far, and exit if we hit
                # the limit
                if nflows_submitted >= self.nflows_max:
                    print("Maxium number of workflow runs hit. Shutting down worker.")
                    break

                # Run and submit flows
                flows_new = await agent.get_and_submit_flow_runs()
                nflows_new = len(flows_new)

                # Keep track of the new runs we submit
                if nflows_new:
                    nflows_submitted += nflows_new
                    print(f"Found and submitted {nflows_new} new workflow runs.")

                # If our last query gave zero workflows submitted, then there's
                # a chance our queue is empty and we should shut down.
                # check the length of the queue and while it is empty, we want to
                # loop. The exception of looping endlessly is if we want the worker
                # to shutdown instead.
                # This is a special condition where we may want to close the
                # worker if the queue stays empty
                if nflows_new == 0 and self.close_on_empty_queue:
                    remaining = await self._get_remaining_runs()
                    if remaining == 0:
                        print("The queue is empty. Shutting down worker.")
                        break

                # Sleep until the next check
                await anyio.sleep(PREFECT_AGENT_QUERY_INTERVAL.value())
                # BUG: using anyio.sleep doesn't shut down the worker properly
                # and it just ignores the ctrl+c command.
                # This line works but doesn't stream logs while sleeping:
                # time.sleep(PREFECT_AGENT_QUERY_INTERVAL.value())
                # Therefore this code is not used:
                # try:
                #   << all code in the while loop >>
                # except KeyboardInterrupt:
                #     print(
                #         "Recieved keyboard signal to stop. Shutting down worker."
                #     )
                #     break

        # Delete the queue now that we are done with it
        print("Deleting queue and cleaning up...")
        await self._delete_queue()
        print("Successfully closed worker.")

    @cached_property
    @async_to_sync
    async def queue_id(self) -> str:
        """
        Grabs the queue id from the prefect database if it exists, and
        if not, creates the queue and then returns the id.

        This is a synchronous and cached version of `_get_queue_id` and
        this is the preferred method to use for beginners.
        """
        return await self._get_queue_id()

    async def _get_queue_id(self) -> str:
        """
        Grabs the queue id from the prefect database if it exists, and
        if not, creates the queue and then returns the id.

        This is an asynchronous method and should only be used when within
        other async methods. Beginners should instead use the `queue_id`
        property.
        """

        try:
            async with get_client() as client:
                response = await client.read_work_queue_by_name(
                    name=self.queue_name,
                )

            # There should only be one queue for each name, so this is returned
            # as a single value
            queue_id = str(response.id)

        # A 404 error will be raised when the queue doesn't exist
        except PrefectHTTPStatusError:
            # If this is the first time accessing the deployment id, we will need
            # to create the queue_id
            queue_id = await self._create_queue()

        return queue_id

    async def _create_queue(self) -> str:
        """
        Registers this queue to the prefect database.

        This method should not be called directly. It will be called by
        other methods when appropriate
        """

        async with get_client() as client:
            response = await client.create_work_queue(
                name=self.queue_name,
                tags=["simmate"],
            )

            # create_work_queue does not accept an input for concurrency, so
            # I need to update this in a 2nd call to the api
            await client.update_work_queue(
                id=str(response),
                concurrency_limit=self.concurrency_limit,
            )

        queue_id = str(response)

        return queue_id

    @staticmethod
    async def _get_remaining_runs() -> int:
        """
        Returns a count of the number of remaining workflow runs.

        This is used to determine when workers should shut down within the `run`
        method, and this method should not be called by users.
        """

        # get_runs_in_work_queue cant look past the concurrency limit
        # so we have to query for the allowed flow runs directly

        async with get_client() as client:
            response = await client.read_flow_runs(
                # BUG: Why won't this work...?
                # flow_filter=FlowFilter(
                #     tags={"all_": ["simmate"]},
                # ),
                flow_run_filter=FlowRunFilter(
                    state={"type": {"any_": ["SCHEDULED", "PENDING"]}}
                ),
            )
        return len(response)

    async def _delete_queue(self):
        """
        Deletes the work queue associated with this worker.

        This method is called automatically at the end of `run` and
        should not be called by users.
        """

        queue_id = await self._get_queue_id()

        async with get_client() as client:
            response = await client.delete_work_queue_by_id(
                id=queue_id,
            )



# -*- coding: utf-8 -*-

import os

from simmate.toolkit import Structure
from simmate.workflow_engine import Workflow
from simmate.workflow_engine.common_tasks import (
    load_input_and_register,
    parse_multi_command,
)

from simmate.calculators.vasp.workflows.diffusion.utilities import (
    get_migration_images_from_endpoints,
)
from simmate.calculators.vasp.workflows.relaxation.neb_endpoint import (
    Relaxation__Vasp__NebEndpoint,
)
from simmate.calculators.vasp.workflows.diffusion.neb_from_images import (
    Diffusion__Vasp__NebFromImages,
)
from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
)


class Diffusion__Vasp__NebFromEndpoints(Workflow):
    """
    Runs a full diffusion analysis from a start and end supercell using NEB.

    The input start/end structures will be relaxed using the relaxation/neb-endpoint
    workflow and then interpolated to generate 7 images. These 7 images are
    then relaxed using NEB within the diffusion/from-images workflow.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/neb-endpoint (for both start and end supercells)
        - a mini task that interpolated relaxed endpoints and makes images
        - diffusion/from-images

    If you are running this workflow via the command-line, you can run this
    with...

    ``` bash
    simmate workflows run diffusion/all-paths -c "cmd1; cmd2" --supercell_start example1.cif --supercell_end example2.cif
    ```

    Note, the `-c` here is very important! Here we are passing three commands
    separated by semicolons. Each command is passed to a specific workflow call:

        - cmd1 --> used for endpoint supercell relaxations
        - cmd2 --> used for NEB

    Thus, you can scale your resources for each step. Here's a full -c option:

    ```
    -c "mpirun -n 12 vasp_std > vasp.out; mpirun -n 70 vasp_std > vasp.out"
    ```
    """

    database_table = MITDiffusionAnalysis

    description_doc_short = "runs NEB using two endpoint structures as input"

    @classmethod
    def run_config(
        cls,
        supercell_start: Structure,
        supercell_end: Structure,
        directory: str = None,
        source: dict = None,
        command: str = None,
        # This helps link results to a higher-level table.
        diffusion_analysis_id: int = None,
        is_restart: bool = False,
    ):

        # command list expects three subcommands:
        #   command_bulk, command_supercell, and command_neb
        subcommands = parse_multi_command(
            command,
            commands_out=["command_supercell", "command_neb"],
        )
        # I separate these out because each calculation is a very different scale.
        # For example, you may want to run the bulk relaxation on 10 cores, the
        # supercell on 50, and the NEB on 200. Even though more cores are available,
        # running smaller calculation on more cores could slow down the calc.

        # Load our input and make a base directory for all other workflows to run
        # within for us.
        parameters_cleaned = load_input_and_register(
            supercell_start=supercell_start,
            supercell_end=supercell_end,
            source=source,
            directory=directory,
            command=command,
            diffusion_analysis_id=diffusion_analysis_id,
            is_restart=is_restart,
        )

        # Relax the starting supercell structure
        endpoint_start_state = Relaxation__Vasp__NebEndpoint.run(
            structure=supercell_start,
            command=subcommands["command_supercell"],
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + f"{Relaxation__Vasp__NebEndpoint.name_full}.start",
            is_restart=is_restart,
        )

        # Relax the ending supercell structure
        endpoint_end_state = Relaxation__Vasp__NebEndpoint.run(
            structure=supercell_end,
            command=subcommands["command_supercell"],
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + f"{Relaxation__Vasp__NebEndpoint.name_full}.end",
            is_restart=is_restart,
        )

        # wait for the endpoint relaxations to finish
        endpoint_start_result = endpoint_start_state.result()
        endpoint_end_result = endpoint_end_state.result()

        # interpolate / empirically relax the endpoint structures
        images = get_migration_images_from_endpoints(
            supercell_start={
                "database_table": Relaxation__Vasp__NebEndpoint.database_table.__name__,
                "directory": endpoint_start_result["directory"],
                "structure_field": "structure_final",
            },
            supercell_end={
                "database_table": Relaxation__Vasp__NebEndpoint.database_table.__name__,
                "directory": endpoint_end_result["directory"],
                "structure_field": "structure_final",
            },
        ).result()

        # Run NEB on this set of images
        neb_state = Diffusion__Vasp__NebFromImages.run(
            migration_images=images,
            command=subcommands["command_neb"],
            source=parameters_cleaned["source"],
            directory=parameters_cleaned["directory"],
            diffusion_analysis_id=parameters_cleaned["diffusion_analysis_id"],
            is_restart=is_restart,
        )



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class PointGroup(ErrorHandler):
    """
    Fixes an error where VASP does not have the symmetry group operations
    available. To fix this, we simply need to turn symmetry off.
    """

    # run this while the VASP calculation is still going
    is_monitor = True

    # we assume that we are checking the vasp.out file
    filename_to_check = "vasp.out"

    # These are the error messages that we are looking for in the file
    possible_error_messages = ["group operation missing"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # turn off symmetry
        incar["ISYM"] = 0
        correction = "switched ISYM to 0"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

import pytest

from simmate.workflow_engine import Workflow, task
from simmate.website.test_app.models import TestStructureCalculation


@task
def dummy_task_1(a):
    return 1


@task
def dummy_task_2(a):
    return 2


class DummyProject__DummyCaclulator__DummyPreset(Workflow):
    """
    Minimal example of a workflow
    """

    database_table = TestStructureCalculation

    @staticmethod
    def run_config(source=None, structure=None, **kwargs):
        x = dummy_task_1(source)
        y = dummy_task_2(structure)
        return x + y


# copy to variable for shorthand use
DummyFlow = DummyProject__DummyCaclulator__DummyPreset


@pytest.mark.prefect_db
def test_workflow():
    # Run the workflow just like you would for the base Prefect class
    flow = DummyFlow.to_prefect_flow()
    state = flow(return_state=True)
    assert state.is_completed()
    assert state.result() == 3

    # Same exact thing but using higher-level method
    state = DummyFlow.run()
    assert state.is_completed()
    assert state.result() == 3

    # testing naming conventions
    assert DummyFlow.name_full == "dummy-project.dummy-caclulator.dummy-preset"
    assert DummyFlow.name_project == "dummy-project"
    assert DummyFlow.name_calculator == "dummy-caclulator"
    assert DummyFlow.name_preset == "dummy-preset"

    # testing class properties
    assert DummyFlow.description_doc == DummyFlow.__doc__
    assert DummyFlow.description_doc.strip() == "Minimal example of a workflow"
    assert DummyFlow.parameter_names == ["kwargs", "source", "structure"]
    assert DummyFlow.parameters_to_register == [
        "prefect_flow_run_id",
        "source",
        "structure",
    ]
    DummyFlow.show_parameters()  # a print statment w. nothing else to check


# @pytest.mark.prefect_db
# @pytest.mark.django_db
# def test_workflow_cloud(mocker, sample_structures):

#     # test cloud properties
#     deployment_id = DummyFlow.deployment_id
#     assert isinstance(deployment_id, str)
#     # we dont check the actual value bc its randomly generated

#     n = DummyFlow.nflows_submitted
#     assert isinstance(n, int)

#     # to test serialization of input parameters we grab a toolkit object
#     structure = sample_structures["C_mp-48_primitive"]

#     # Run the workflow through prefect cloud
#     flow_id = DummyFlow.run_cloud(structure=structure)
#     assert isinstance(flow_id, str)


def test_serialize_parameters():
    class TestParameter1:
        def to_dict(self):
            return {}

    class TestParameter2:
        """requires cloudpickle to serialize"""

        a = 123

    parameters = dict(
        a=TestParameter1(),
        b=TestParameter2(),
    )
    Workflow._serialize_parameters(**parameters)


def test_deserialize_parameters(mocker):

    # -------
    # We don't want to actually call these methods, but just ensure that they
    # have been called.
    from simmate.toolkit import Structure
    from simmate.toolkit.diffusion import MigrationHop

    mocker.patch.object(
        Structure,
        "from_dynamic",
    )
    mocker.patch.object(
        MigrationHop,
        "from_dynamic",
    )
    # -------

    example_parameters = {
        "migration_hop": None,
        "supercell_start": None,
        "supercell_end": None,
        "structures": "None; None; None",
    }

    Workflow._deserialize_parameters(**example_parameters)



# -*- coding: utf-8 -*-

"""

This file is for pulling AFLOW data into the Simmate database. 

Materials Cloud looks to be composed of smaller databases, which makes this tricky.
After reading through their website, I don't see an easy way to access all of their
data. It does look like there is an easier OPTIMADE endpoint here though:
    https://www.optimade.org/providers-dashboard/providers/mcloud.html
OPTIMADE isn't super clear on how to query these subdatabases though. I need to 
do some more digging. I can get a single page of structures from this REST endpoint
though:
    https://aiida.materialscloud.org/3dd/optimade/structures
Read more at:
    https://petstore.swagger.io/?url=https://raw.githubusercontent.com/Materials-Consortia/OPTIMADE/v1.0.0/schemas/openapi_schema.json#/

Another option is using pymatgen's OPTIMADE class:
    https://github.com/materialsproject/pymatgen/blob/v2022.0.8/pymatgen/ext/optimade.py

"""



# -*- coding: utf-8 -*-

##############################################################################

import numpy
import itertools

from simmate.toolkit.validators.base import Validator

##############################################################################


class SiteDistance(Validator):
    def __init__(self, distance_cutoff):

        self.distance_cutoff = distance_cutoff

    def check_structure(self, structure):

        # NEED A BETTER WAY TO DO THIS!
        # but to check the min distance of sites, I need to ignore the diangonal (which is also made of zeros)
        dm = structure.distance_matrix
        nonzeros = dm[numpy.nonzero(dm)]
        min_distance = nonzeros.min()

        # check if min is larger than a tolerance
        check = min_distance >= self.distance_cutoff

        return check


##############################################################################


class SiteDistanceMatrix(Validator):
    def __init__(
        self,
        composition,
        radius_method="ionic",
        packing_factor=0.5,  # options unique to this class
    ):

        # save inputs for reference
        self.composition = composition
        self.radius_method = radius_method
        self.packing_factor = packing_factor

        # using our base predictor, make the distance matrix
        self.element_distance_matrix = composition.distance_matrix_estimate(
            radius_method,
            packing_factor,
        )

        #!!! BUG - If I use the ionic radius method, I should convert compositions and structures to the oxidation-state-decorated structures!
        # Logical errors will happen because composition changes - for example, Mg4Si8O12 has Mg2+4 Si4-2 Si4+6 O2-12.
        # This gives a 4x4 distance matrix, which won't match up to the expected 3x3 of a non-decorated structure.
        # For this reason, I have the default radius method set to atomic. IONIC SHOULD NOT BE SELECTED AT THE MOMENT!!!!
        #!!! quick BUG fix - this is not the best fix though (see comment above)
        # With this fix, elements predicted to have different oxidation states
        # will be restricted to only their smallest type.
        # For example, if Nitrogen is predicted to have Specie N3+, Specie N5+, Specie N3-
        # then a column is made for each of these species in the distnace matrix.
        # But because I don't convert the structure to an oxidation-state-decorated structure
        # in the check_structure() function below, all Nitrogens are checked for all of these
        # different species. Thus we are doing repeat checks and the smallest radius specie is
        # the only one having a real effect (as it sets the true min distance)
        if radius_method == "ionic":
            self.composition = composition.add_charges_from_oxi_state_guesses(
                max_sites=-1
            )

    def check_structure(self, structure):
        # now using the matrix above, we need to look at the structure
        # and determine if there are any distances that are below matrix limits
        # to save time, we want to iterate through each site and see that the site meets all requirements
        # as soon as any requirement is failed for any site, the whole loop ends
        #!!! This function is slow.
        # Maybe instead of making all comparisons, we can only look at nearest neighbors

        # we get a massive speedup if we make this matrix upfront
        # for extremely large structures with many sites, this might not be the case though - I still need to test for that
        dm = structure.distance_matrix

        # go through each element-element combo and check distances
        for i1, element1 in enumerate(self.composition):
            element1_indicies = structure.indices_from_symbol(element1.symbol)
            for i2, element2 in enumerate(self.composition):
                # grab the cutoff distance for these two specie/element types
                dist_cutoff = self.element_distance_matrix[i1][i2]
                element2_indicies = structure.indices_from_symbol(element2.symbol)
                combos = itertools.product(element1_indicies, element2_indicies)
                for s1, s2 in combos:
                    # skip if we are looking at the same site
                    if s1 == s2:
                        continue
                    # grab the periodic sites
                    # NOTE: sites here are PeriodicSite objects, not Site objects
                    # This is extremely important because we want nearest image distance, which might not be in an adjacent cell!
                    # site1 = structure[s1]
                    # site2 = structure[s2]
                    # Now get the distance between the two periodic sites
                    # distance = site1.distance(site2)
                    distance = dm[s1][s2]
                    # compare the distance to the cutoff matrix
                    if distance < dist_cutoff:
                        # one False is enough to stop - end the whole function
                        return False
        # the function will only reach this point if all distance criteria are met
        return True


##############################################################################

# # distance cutoff matrix based on USPEX (used inputgenerator)
# # read this matrix with the following:
# #    Mg Si O
# # Mg
# # Si
# # O
# dist_cut_matrix = numpy.array([[1.13, 1.06, 0.96],
#                                [1.06, 0.98, 0.89],
#                                [0.96, 0.89, 0.79]])

##############################################################################



# -*- coding: utf-8 -*-

import os
import sys
import requests

import simmate


def get_conda_env() -> str:
    """
    Grab the name of the activate conda environment and returns it as a string.
    If there is no env, then an empty string is returned.
    """
    # Check the list of python paths and grab the first path that has "envs" in it.
    # Assume we don't have a conda env until proven otherwise
    env_name = ""
    for path in sys.path:
        if "envs" in path:
            # split the path into individual folder names (os.sep gives / or \\)
            folders = path.split(os.sep)
            # the conda env name will the name immediately after the /envs.
            # example path is '/home/jacksund/anaconda3/envs/simmate_dev/lib/python3.10'
            # where we want the name simmate_dev here.
            env_name = folders[folders.index("envs") + 1]
            # once we have found this, we can exit the loop
            break

    return env_name


def get_doc_from_readme(file: str) -> str:
    """
    Loads the docstring from a README.md file in the same directory.

    This is commonly used in __init__.py files because we like having our
    documentation isolated (so that github renders it).

    To use, simply pass the file property:

    ``` python
    from simmate.utilities import get_doc_from_readme

    __doc__ = get_doc_from_readme(__file__)
    ```

    This is an alternative to using "include" in rst files, which
    [pdoc recommends](https://pdoc.dev/docs/pdoc.html#include-markdown-files).
    We prefer this utility because it allows Spyder to load the docs -- although
    it's slower in production (bc of opening/closing files).
    """

    # We assume the file is in the same directory and named "README.rst"
    file_directory = os.path.dirname(os.path.abspath(file))
    with open(
        os.path.join(file_directory, "README.md"),
        encoding="utf-8",
    ) as doc_file:
        doc = doc_file.read()
    return doc


def get_latest_version() -> str:
    """
    Looks at the jacks/simmate repo and grabs the latest release version.
    """
    # Access the data via a web request
    response = requests.get(
        "https://api.github.com/repos/jacksund/simmate/releases/latest"
    )

    # load the version from the json response
    # [1:] simply removes the first letter "v" from something like "v1.2.3"
    latest_version = response.json()["tag_name"][1:]
    return latest_version


def check_if_using_latest_version(current_version=simmate.__version__):
    """
    Checks if there's a newer version by looking at the latest release on Github
    and comparing it to the currently installed version
    """
    latest_version = get_latest_version()

    if current_version != latest_version:
        print(
            "WARNING: There is a new version of Simmate available. \n"
            f"You are currently using v{current_version} while v{latest_version} "
            "is the latest. To update, you should first check what has been "
            "changed at https://github.com/jacksund/simmate/blob/main/CHANGELOG.md. "
            "If you would like to use these changes, you can run the follwing "
            "command in the terminal (with your desired conda environment "
            "active): `conda update --all`"
        )


def get_chemical_subsystems(chemical_system: str):
    """
    Given a chemical system, this returns all chemical systems that are also
    contained within it.

    For example, "Y-C" would return ["Y", "C", "C-Y"]. Note that the returned
    list has elements of a given system in alphabetical order (i.e. it gives
    "C-Y" and not "Y-C")

    #### Parameters

    - `chemical_system`:
        A chemical system of elements. Elements must be separated by dashes (-)

    #### Returns

    - `subsystems`:
        A list of chemical systems that make up the input chemical system.
    """

    # TODO: this code may be better located elsewhere. Maybe even as a method for
    # the Composition class or alternatively as a ChemicalSystem class.

    # I convert the system to a composition where the number of atoms dont
    # apply here. (e.g. "Ca-N" --> "Ca1 N1")
    from simmate.toolkit import Composition

    composition = Composition(chemical_system.replace("-", ""))

    return composition.chemical_subsystems



# -*- coding: utf-8 -*-

import os
import pytest
from simmate.workflow_engine.error_handler import ErrorHandler


class CheckREADME(ErrorHandler):
    """
    This checks if simmate is in located in this file (which will always be true).

    This is just to check the default check() method for the ErrorHandler class.
    """

    filename_to_check = __file__
    possible_error_messages = ["simmate"]

    def correct(self, directory):
        return "ExampleCorrection"


class IncorrectHandler(ErrorHandler):
    # filename_to_check  # <-- incorrectly missing this field
    # possible_error_messages <-- incorrectly missing this field

    def correct(self, directory):
        return "ExampleCorrection"


def test_error_handler():

    directory = os.path.dirname(__file__)

    # test basic use
    handler = CheckREADME()
    assert handler.check(directory=directory)
    assert handler.correct(directory=directory) == "ExampleCorrection"

    # confirm Exception when settings are improperly set
    handler = IncorrectHandler()
    with pytest.raises(Exception):
        handler.check(directory=directory)

    # This line tests nothing, but simply covers the abstract method's pass statement
    ErrorHandler.correct(None, None)



# -*- coding: utf-8 -*-

# For adding new selection methods, DEAP is a good reference:
#   https://github.com/DEAP/deap/blob/master/deap/tools/selection.py
#   https://deap.readthedocs.io/en/master/api/tools.html
#
# Their selection methods take a list of individuals that each have a .fitness
# attribute. Instead, I pass individuals here as a pandas dataframe and have
# them specify a "fitness_column"
#
# Another good reference is Wikipedia
#   https://en.wikipedia.org/wiki/Selection_(genetic_algorithm)

# NOTE: I assume lower fitness is better! This may change in the future because
# other selection methods throw issues with lower=better and even more issues
# when we have negative values


class Selector:
    def __init__(
        self,
    ):
        # add any unchanging variables or settings here
        pass

    def select(self, nselect, individuals, fitness_column):
        # The select method should take in two arguments:
        # individuals = a pandas dataframe with columns "id" and "fitness"
        # n = the number of individuals that we want to choose from those fitnesses
        # The select method should then return a list/tuple of the indicies
        # of the fitnesses that were selected
        pass


class TruncatedSelection:
    """
    Truncated selection limits the parent selection to the top X% of individuals,
    and then within this top section, every individual has an equal chance of
    being selected.
    """

    def __init__(self, percentile=0.05, ntruncate_min=5, allow_duplicate=True):

        # aim to grab this percent of individuals
        self.percentile = percentile

        # at minimum, make sure we can select from this number of best individuals
        self.ntruncate_min = ntruncate_min

        # whether we can select the same individual more than once
        self.allow_duplicate = allow_duplicate

    def select(self, nselect, individuals, fitness_column):

        # truncate the population to those with energies in the lowest X%. This
        # value should be greater or equal to our minimum set above
        ntruncate = int(len(individuals) * self.percentile)
        if ntruncate < self.ntruncate_min:
            ntruncate = self.ntruncate_min

        # limit our potential selection to the top X% and then randomly select
        # the correct number of parents
        df_truncated = individuals.nsmallest(ntruncate, fitness_column)
        df_parents = df_truncated.sample(nselect, replace=self.allow_duplicate)

        # return the list of indexes to be selected
        return df_parents



# -*- coding: utf-8 -*-

import os
import json

from simmate.toolkit import Structure
from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class RealOptlay(ErrorHandler):
    """
    This handler addresses a series of warning messages that each have the same
    attempted fixes. There are a series of fixes that can be attempted, so we
    keep a running log of how many times this handler is called.
    """

    # NOTE: I call this RealOptlay but don't have a good understanding of what
    # the error exactly is. This is simply a conversion from Custodian of the
    # following errors: ["rspher", "real_optlay", "nicht_konv"]

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = [
        "ERROR RSPHER",
        "REAL_OPTLAY: internal error",
        "REAL_OPT: internal ERROR",
        "ERROR: SBESSELITER : nicht konvergent",
    ]

    # Number of atoms in a lattice to consider this a large cell. This affects
    # how we treat the error and correct it.
    natoms_large_cell = 100

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # load the error-count file if it exists
        error_count_filename = os.path.join(directory, "simmate_error_counts.json")
        if os.path.exists(error_count_filename):
            with open(error_count_filename) as error_count_file:
                error_counts = json.load(error_count_file)
        # otherwise we are starting with an empty dictionary
        else:
            error_counts = {}

        # The fix is based on the number of times we've already tried to
        # fix this. So let's make sure it's in our error_count dictionary.
        # If it isn't there yet, set the count to 0 and we'll update it below.
        error_counts["real_optlay"] = error_counts.get("real_optlay", 0)

        poscar_filename = os.path.join(directory, "POSCAR")
        structure = Structure.from_file(poscar_filename)

        if structure.num_sites < self.natoms_large_cell:
            incar["LREAL"] = False
            correction = "set LREAL to False"

        else:
            # for large supercell, try an in-between option LREAL = True
            # prior to LREAL = False
            if error_counts["real_optlay"] == 0:
                # use real space projectors generated by pot
                incar["LREAL"] = True
                correction = "set LREAL to True"
            elif error_counts["real_optlay"] == 1:
                incar["LREAL"] = False
                correction = "set LREAL to False"
            # increase our attempt count
            error_counts["real_optlay"] += 1

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        # rewrite the new error count file
        with open(error_count_filename, "w") as file:
            json.dump(error_counts, file)

        # now return the correction made for logging
        return correction



# -*- coding: utf-8 -*-

from pymatgen.io.vasp.outputs import Vasprun
from simmate.database.base_data_types import table_column, Calculation


class CustomizedCalculation(Calculation):
    """
    A customized calculation is when the corresponding workflow has been
    fundamentally changed -- meaning critical settings for the workflow were
    updated. For example, we may want to run a workflow that uses different
    convergence critria or even add a correction setting (e.g. a dispersion
    correction). Simmate aims to store this data separately from the main
    workflow's result table, so this table holds the results from altered runs.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    workflow_base = table_column.CharField(blank=True, null=True, max_length=75)
    """
    The name of the workflow that is updated. (e.g. "relaxation/matproj")
    """

    input_parameters = table_column.JSONField(blank=True, null=True)
    """
    The attributes updated on the `base_workflow` before the flow was ran
    """

    updated_settings = table_column.JSONField(blank=True, null=True)
    """
    The attributes updated on the `base_workflow` before the flow was ran
    """

    data = table_column.JSONField(blank=True, null=True)
    """
    The final data from the calculation. Because there are a variety of possible
    results, all data is stored as JSON.
    """

    def update_from_vasp_run(
        self,
        vasprun: Vasprun,
        corrections: list,
        directory: str,
    ):
        # Takes a pymatgen VaspRun object, which is what's typically returned
        # from a simmate VaspTask.run() call.

        # rather than cherry-pick data, we just save everything as a dictionary
        self.data = vasprun.as_dict()

        # lastly, we also want to save the corrections made and directory it ran in
        self.corrections = corrections
        self.directory = directory

        # Now we have the relaxation data all loaded and can save it to the database
        self.save()



# -*- coding: utf-8 -*-

# !!! In the future, I think this should be a "converter" class that loads data
# into a more generic DftResults class -- so one that is more generically for
# storing electronic and ionic steps across different calculators. I think it
# would be useful to have a ElectronicStep and IonicStep classes.

import numpy


class Oszicar:
    """
    Reads all data from an OSZICAR file. This includes data on each electronic
    and ionic step of the calculation.

    Note that this class is only used when the speed of file-reading is important.
    If you are trying to analyze your VASP run, you should instead use the VaspXML
    output, which gives all of the information already in the OSZICAR and more!

    To help with understanding the OSZICAR file, you can also look here:
        https://www.vasp.at/wiki/index.php/OSZICAR
    """

    def __init__(self, filename="OSZICAR"):

        # Make an empty list to store all of the ionic steps. While relaxations
        # have many ionic steps, note that a static energy calculation will only
        # ever have one step. Either way, we still store the steps in a list.
        self.ionic_steps = []

        # open the file, read it's contents, and close immediately
        with open(filename) as file:
            lines = file.readlines()
        # Empty lines cause trouble so we remove those before parsing any data.
        lines = [line for line in lines if line.strip()]

        # now let's iterate through the contents and parse the data. We also need
        # to check the line number below, so we use the index too
        for line_number, line in enumerate(lines):

            # if "N" is in the line, that means we have the start of a new ionic step
            if "N" in line:

                # let's start with an empty list to record all of the electronic
                # steps. Note this will also reset the list every time we start
                # a new ionic step.
                electronic_steps = []
            # if "F" is in the line, that means we have hit the end of an ionic step.
            # We also need to account for when the calculation is not finished and
            # in the middle of an ionic step. Therefore we also check if this is the final
            # line of the file.
            elif "F" in line:

                # parse this line to pull out the ionic_step information
                # The formatting is whacky, so I decided to remove all variable
                # names, leaving only a list of values. I then split these values
                # into a list that I can more easily work with.
                values = (
                    line.strip()
                    .replace("T= ", "")
                    .replace("E= ", "")
                    .replace("F= ", "")
                    .replace("E0= ", "")
                    .replace("d E =", "")
                    .replace("mag=", "")
                    .replace("EK=", "")
                    .replace("SP=", "")
                    .replace("SK=", "")
                    .split()
                )

                # convert the list of strings to float values. We skip the first
                # value which is just the ionic_step_number that we don't need
                values = [float(value) for value in values[1:]]

                # Now there are three formats that we need to account for, which
                # each have a different amount of information given, and we can
                # figure out the type easy based on the length of our list of
                # values. These are...
                #   (1) non-magnetic calculations (F, E0, dE)
                #   (2) molecular dynamics calculations (T, E, F, E0, EK, SP, SK)
                # Further, magnetic calculations will have another value added
                # on at the end (mag). We account for this below too.

                # check if we have a simple energy calculation
                if len(values) in [3, 4]:
                    ionic_step = {
                        "energy": values[0],  # F (total_free_energy)
                        "energy_sigma_zero": values[1],  # E0
                        "energy_change": values[2],  # dE
                        "electronic_steps": electronic_steps,
                    }
                # check if we have a molecular dynamics calculation
                elif len(values) in [7, 8]:
                    ionic_step = {
                        "temperature": values[0],  # T
                        "energy_gibbs": values[1],  # E (total energy, kinetic, nose)
                        "energy": values[2],  # F (total_free_energy)
                        "energy_sigma_zero": values[3],  # E0
                        "energy_kinetic": values[4],  # EK
                        "energy_potential_nose_thermostat": values[5],  # SP
                        "energy_kinetic_nose_thermostat": values[6],  # SK
                        "electronic_steps": electronic_steps,
                    }

                # otherwise something went wrong
                else:
                    raise Exception(
                        "Electronic step had unexpected data. Failed to parse."
                    )

                # If the length of the values list is 4 or 8, then that means
                # we also have a "mag" value attatched at the end of our values
                if len(values) in [4, 8]:
                    ionic_step.update({"magnetic": values[-1]})

                # for the very first electronic step, VASP provides an energy-change
                # value, which they set equal to the energy itself. This is misleading
                # in analysis so we remove it.
                # As an example of why this is important, this will cause problems
                # in the PotimErrorHandler when the original structure is a poor
                # guess and positive energy. The handler will mistakenly think
                # the energy changes are getting worse -- which is not the case
                # because we only look at the first step!
                if not self.ionic_steps:
                    ionic_step["energy_change"] = numpy.NaN
                # take whatever ionic_step that was made above and add it to our
                # results list!
                self.ionic_steps.append(ionic_step)
            # Otherwise this line is an electronic step
            else:

                # electronic step lines are pretty easy because they are just
                # a row in a table of electronic steps. First let's split
                # the values into and list and then convert them into floats.
                # Note we remove the first two values because these are the
                # scheme used (i.e. DAV, RMM, or CG) and electronic step number
                # which we don't need. (I grab the scheme below though)
                # BUG: sometimes VASP prints a value like '-0.33328-312' which
                # can't be converted to a float. I need a try/except here to
                # catch this case and return a numpy.NAN instead.
                def try_float(value):
                    try:
                        return float(value)
                    except:
                        return numpy.NaN

                values = [try_float(value) for value in line.split()[2:]]

                # now load the data into a dictionary for verbosity
                # Note that I change the VASP headers to more verbose names so
                # the user can instantly see their meaning. The original headers
                # are shown as comments next to each.
                electronic_step = {
                    # The scheme is set by IALGO and is the first thing written
                    # in the line. Rather than grab it above, I just do it here.
                    "scheme": line.split()[0].replace(":", ""),
                    "energy": values[0],  # E
                    "energy_change": values[1],  # dE
                    "band_structure_energy_change": values[2],  # d eps
                    "nhamiltonians": values[3],  # ncg
                    "wavefunctions_residuum_norm": values[4],  # rms
                    # Note not all electronic steps have this value so we need to doublecheck
                    "charge_density_change": values[5]
                    if len(values) == 6
                    else None,  # rms(c)
                }

                # add the electronic step to our results!
                electronic_steps.append(electronic_step)

                # SPECIAL CASE:
                # if this is the final line and isn't an ionic step summary,
                # then that means we are in the middle of the ionic step and
                # are reading an incomplete file. We still want all the
                # electronic steps available in the final class so we store
                # the incomplete ionic step here. This is useful for ErrorHandlers.
                # We use numpy.NaN instead of None so errors aren't through
                # elsewhere that expect a float value.
                if line_number == len(lines) - 1:
                    # BUG: we don't know what the data should be here (see the
                    # three formats discussed above). For now, we assume
                    # the first format as it is most common. We are filling this
                    # data with "None" values anyways so it shouldn't matter.
                    ionic_step = {
                        "energy": numpy.NaN,  # F (total_free_energy)
                        "energy_sigma_zero": numpy.NaN,  # E0
                        "energy_change": numpy.NaN,  # dE
                        "electronic_steps": electronic_steps,
                    }
                    self.ionic_steps.append(ionic_step)

    def all_electronic_step_energies(self, ionic_step_number):
        # TODO: move to DftCalc/IonicStep/ElectronicStep class
        pass

    def all_ionic_step_energies(self):
        # TODO: move to DftCalc/IonicStep/ElectronicStep class
        pass

    @property
    def energy_final(self):
        # TODO: move to DftCalc/IonicStep/ElectronicStep class
        return self.ionic_steps[-1]["energy_sigma_zero"]



# -*- coding: utf-8 -*-

import math


class Kpoints:
    """
    This class takes a structure along with a KptGrid or KptPath object and
    then generates a VASP formatted KPOINTS file for you. Alternatively,
    you can provide a structre and a KPOINTS file to get back a KptGrid or
    KptPath object. This behaves as a converter class, just like the POSCAR class.
    """

    @staticmethod
    def write_file(
        # structure to base off of. this will be passed through the kpoint_method
        structure,
        # either a KptGrid or KptPath object. TODO - KptGrid is just a float (density)
        kpoint_method,  # !!! JUST A FLOAT VALUE FOR NOW
        filename="KPOINTS",
        comment=None,
    ):

        # let's have a list where we keep track of list of lines that we
        # will join at the end
        lines = []

        # first line is just a comment. If the user didn't supply one, I will
        # just use "Generated by Simmate with [name of kpoint_method]"
        if not comment:
            comment = f"Generated by Simmate at {str(kpoint_method)} kpts per Ang^-1"
        lines.append(comment)

        # check to see if kpoint_method is a KptGrid object
        # TODO - for now this is just a float value as I didn't write the class
        # NOTE - while we suport this, the prefered method is to set kpt_density
        # through the KSPACING tag in the INCAR.
        # BUG - make sure integers are converted to floats, even though a kpt
        # density of 1 or greater should never be used, we should still be
        # ready for it.
        if isinstance(kpoint_method, float):

            # append 0 to the second line which tells VASP that the grid should
            # be generated automatically based on the lines below it
            lines.append("0")

            # The next line indicates whether we use a Gamma-centered or a
            # Monkhorst-Pack grid. We can determine this based off of what
            # type of lattice the structure has:
            #   Gamma centered for hexagonal cells
            #   Otherwise use a Monkhorst-Pack grid
            # TODO -- should this be inside the KptGrid class? I not sure if
            # this is universal or unique to VASP.
            if structure.lattice.is_hexagonal():
                lines.append("Gamma")
            else:
                lines.append("Monkhorst-Pack")

            # Now write out the kpt grid in the form "N1 N2 N3"
            # TODO - this code should be implemented in the KptGrid class
            # with a method that I then reformat. This also includes whether
            # this is a singleline or not as well as if there is a shift
            # applied at the end
            # iterate through the reciprocal lattice vectors and find the values
            # of N to use for each.
            n_list = []
            for vector in structure.lattice.reciprocal_lattice.abc:
                # check how many subdivisions we need to cut this reciprocal
                # lattice vector into in order to get X kpts per Angstrom^-1
                # I want to round up to the nearst integer, which requires
                # the math.ceil function (as opposed to the "//" operator)
                # TODO - reciprocal_density = kpoint_method until I add KptGrid class
                nsubdivisions = math.ceil(vector / kpoint_method)
                n_list.append(nsubdivisions)
            # now combine the list and add it
            lines.append(" ".join(str(n) for n in n_list))

            # TODO - if there is a specified shift in the kpts, add them
            # if kpoint_method.shift:
            #     lines.append(kpoint_method.shift)

        # TODO - I haven't written the KptPath object yet so I can't do this.
        # chek to see if kpoint_method is a KptPath object
        # elif isinstance(kpoint_method, KptPath):
        #     pass

        # Now we can join all of the lines!
        lines = "\n".join(lines) + "\n"

        # Now write the file
        # If I want a separate to_string method, I should move only this code.
        with open(filename, "w") as file:
            file.write(lines)

    # TODO -- I have no reason to read kpt files yet. I likely won't implement
    # this until I add the KptGrid and KptPath classes as well as start doing
    # Band structure calculations
    # @staticmethod
    # def from_file():
    #     pass



# -*- coding: utf-8 -*-

from .base import DatabaseTableFilter

from .symmetry import Spacegroup
from .structure import Structure
from .calculation import Calculation
from .forces import Forces
from .thermodynamics import Thermodynamics

from .static_energy import StaticEnergy
from .relaxation import Relaxation
from .dynamics import DynamicsRun
from .band_structure import BandStructure, BandStructureCalc
from .density_of_states import DensityofStates, DensityofStatesCalc

from .nudged_elastic_band import DiffusionAnalysis



# -*- coding: utf-8 -*-

import numpy as np

from matminer.featurizers.site import CrystalNNFingerprint as cnnf

#!!! is it faster to import these all at once?
from matminer.featurizers.structure import SiteStatsFingerprint as ssf
from matminer.featurizers.structure import RadialDistributionFunction as rdf
from matminer.featurizers.structure import PartialRadialDistributionFunction as prdf
from matminer.featurizers.structure.sites import PartialsSiteStatsFingerprint

from scipy.spatial.distance import cosine


class ASEFingerprint:

    # ASE cites this as a rewrite of the USPEX fingerprint
    # https://gitlab.com/ase/ase/-/blob/master/ase/ga/ofp_comparator.py

    def __init__(self, composition, on_fail, max_fails, initial_structures=[]):

        self.on_fail = on_fail

        # we need the cumulative max_fails list
        # for example, the max_fails=[20,40,10] needs to be converted to [20,60,70]
        # if infinite attempts are allowed, be sure to use float('inf') as the value
        # This list is more efficient to work with in the point_next_step()
        #!!! should I make this numpy for speed below?
        self.max_fails_cum = [sum(max_fails[: x + 1]) for x in range(len(max_fails))]

        # we can assume the user has ASE installed because it is a dependency of PyMatgen
        #!!! it looks like the ase.ga module is actively changing so version may introduce errors
        from ase.ga.ofp_comparator import OFPComparator

        self.featurizer = OFPComparator(  # n_top=None, # atoms to optimize. None is all of them
            # dE=1.0, # energy difference to be considered different (not used here b/c we dont have a calculator)
            # cos_dist_max=5e-3, # cosine distance for structure to be considered different
            # rcut=20., # cutoff radius for the fingerprint
            # binwidth=0.05, # bin width (in Angstrom) over which fingerprints are discretized
            # sigma=0.02, # smearing of the guassian in fingerprint
            # nsigma=4, # distance where smearing is cut off
            # pbc=True, # whether to apply periodic boundry conditions
            # maxdims=None, # not used if we have a 3d crystal
            # recalculate=False # wether to recalculate fingerprints in ASE database
        )

        # we also need to convert pymatgen Structures to ase Atoms below
        #!!! is it faster and more memory efficient to import below?
        from pymatgen.io.ase import AseAtomsAdaptor

        self.adaptor = AseAtomsAdaptor

        # generate the fingerprint for each of the initial input structures
        self.fingerprint_database = []
        for structure in initial_structures:
            # convert the structure to ase Atoms object
            structure_ase = self.adaptor.get_atoms(structure)

            # make the fingerprint
            # ase uses .looks_like() to compare two structures, but we want the lower-level function ._take_fingerprints()
            output = self.featurizer._take_fingerprints(structure_ase)
            # the output is a tuple of...
            # (1) dictionary with keys of each element combo mapped to the fingerprint
            # (2) dictionary with keys of element to the indicies of that element in the Atoms object
            # so I want to make the fingerprint a 1D array so I go through the output and merge these.
            # make a list of each element-element fingerprint
            fingerprints = [output[0][key] for key in output[0]]
            # now combine the fingerprints into a single 1D array
            fingerprint = np.hstack(fingerprints)

            # add to the database
            self.fingerprint_database.append(fingerprint)
        # now make the datebase into a numpy array for speed
        self.fingerprint_database = np.array(self.fingerprint_database)

    def check_structure(self, structure, tolerance=5e-3):

        # convert the structure to ase Atoms object
        structure_ase = self.adaptor.get_atoms(structure)

        # make the fingerprint
        # ase uses .looks_like() to compare two structures, but we want the lower-level function ._take_fingerprints()
        output = self.featurizer._take_fingerprints(structure_ase)
        # the output is a tuple of...
        # (1) dictionary with keys of each element combo mapped to the fingerprint
        # (2) dictionary with keys of element to the indicies of that element in the Atoms object
        # so I want to make the fingerprint a 1D array so I go through the output and merge these.
        # make a list of each element-element fingerprint
        fingerprints = [output[0][key] for key in output[0]]
        # now combine the fingerprints into a single 1D array
        fingerprint1 = np.hstack(fingerprints)

        # we now want to get the distance of this fingerprint relative to all others
        # if the distance is within the specified tolerance, then the structures are too similar - we return false
        # if we make it through all structures and no distance is below the tolerance, we are good to go!
        for fingerprint2 in self.fingerprint_database:
            # we want the cosine distance
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.cosine.html
            # I use the numpy version of this for speed
            distance = cosine(fingerprint1, fingerprint2)
            if distance < tolerance:
                # we can end the whole function as soon as one structure is deemed too similar
                return False
            else:
                continue

        # if we finished the loop above, we can add this structure to the database and return a success
        # this method I use to append is pretty slow for numpy. Is there a better way to do this?
        if self.fingerprint_database.size == 0:
            self.fingerprint_database = np.array([fingerprint1])
        else:
            self.fingerprint_database = np.append(
                self.fingerprint_database, [fingerprint1], axis=0
            )

        return True

    def point_next_step(self, attempt=0):

        # based on the attempt number and the max_fails list, see what stage of on_fail we are on
        for i, attempt_cutoff in enumerate(self.max_fails_cum):
            # if we are below the attempt cutoff, return the corresponding on_fail object
            if attempt <= attempt_cutoff:
                return self.on_fail[i]

        # If we make it through all of the cutoff list, that means we exceeded the maximum attempts allowed
        # In this case, we don't return return an object from the on_fail list
        return False



# -*- coding: utf-8 -*-

"""
This module defines the base class for all workflows in Simmate. When learning 
how use workflows, make sure you have gone through our intro 
[tutorials](https://github.com/jacksund/simmate/tree/main/tutorials). You
can then read through these guides for more features.

It useful to know that our Worklfow class builds Prefect Flows under the 
hood, so having knowledge of Prefect can be useful in advanced cases. For 
advanced use or when building new features, we recommend going through the
prefect tutorials located
[here](https://orion-docs.prefect.io/tutorials/first-steps/).



# Using existing workflows


## Loading a workflow

To import a specific workflow, see the `simmate.workflows` module. For all the
example uses below, we will look at the `static-energy.vasp.matproj` workflow,
which we load with:
    
``` python
from simmate.workflows.all import StaticEnergy__Vasp__Matproj as workflow
```


## Viewing input parameters



## Running a workflow (local)

To run a workflow locally (i.e. directly on your current computer), you can
use the `run` method. This is returns a Prefect `State` object, which allows
you to check if the run completed successfully or not. Then final output of your
workflow run can be accessed using `state.result()`. You can read more about
`State` objects [here](https://orion-docs.prefect.io/concepts/states/), but as
a quick example:

``` python
state = workflow.run(
    structure="NaCl.cif", 
    command="mpirun -n 4 vasp_std > vasp.out",
)

if state.is_completed():  # optional check
    result = state.result()
```

Outside of python, you can also run workflows from the command line, using YAML 
files, or the website interface. These approaches are covered in the intro
tutorials.


## Running a workflow (cloud)

When you want to schedule a workflow to run elsewhere, you must first make sure
you have your computational resources configured. You can then run workflows
using the `run_cloud` method, which returns a Prefect flow run id.

``` python
prefect_flow_run_id = workflow.run_cloud(
    structure="NaCl.cif", 
    command="mpirun -n 4 vasp_std > vasp.out",
)
```

## Accessing results

In addition to using `state.result()` as described above, you can also view
results database through the `database_table` attribute (if one is available).
This returns a Simmate database object for results of ALL runs of this workflow.
Guides for filtering and manulipating the data in this table is covered in
the `simmate.database` module guides. But as an example:

``` python
table = workflow.database_table

# pandas dataframe that you can view in Spyder
df = table.obects.to_dataframe()

# or grab a specific run result and convert to a toolkit object
entry = table.objects.get(prefect_flow_run_id="example-123456")
structure = entry.to_toolkit()
```

<!--
TODO:
    - viewing parameter names
    - accessing results (database_table)
    - run_cloud and other cloud methods
    - different documentation levels (parameters, mini docstring, etc.)
    - finding the same workflow in the website UI (place this )
-->

# Creating new workflows

## Naming workflows (required)

Higher level features such as the website interface require that workflow
names follow a certain format. If you skip this step, your workflows
will fail and cause errors elsewhere.

First, we need to update the workflow name to match Simmate's naming
conventions, which includes:
    1.  The type of analysis the workflow is doing
    2.  The "calculator" (or program) that the workflow uses to run
    3.  A unique name to identify the settings used

Examples for each part would be:
    1. relaxation, static-energy, dynamics, ...
    2. vasp, abinit, qe, deepmd, ...
    3. jacks-test, matproj, quality00, ...

Together, an example workflow names would be:
    - `relaxation.vasp.jacks-test`
    - `static-energy.abinit.matproj`
    - `dynamics.qe.quality00`

When converting this to our workflow name in python, we need to replace
periods with 2 underscores each and convert our words to
[pascal case](https://khalilstemmler.com/blogs/camel-case-snake-case-pascal-case/).
For example, our workflow names become:
    - `Relaxation__Vasp__JacksTest`
    - `StaticEnergy__Abinit__Matproj`
    - `Dynamics__Qe__Quality00`

NOTE: Capitalization is very important here so make sure you double check
your workflow names.

Now let's test this out in python using a similar workflow name:
``` python
from simmate.workflow_engine import Workflow

class Example__Python__MyFavoriteSettings(Workflow):
    pass  # we will build the rest of workflow later

# These names can be long and unfriendly, so it can be nice to
# link them to a variable name for easier access.
my_workflow = Example__Python__MyFavoriteSettings

# Now check that our naming convention works as expected
assert my_workflow.name_full == "example.pure-python.my-favorite-settings"
assert my_workflow.name_project == "example"
assert my_workflow.name_calculator == "python"
assert my_workflow.name_preset == "my-favorite-settings"
```

You now have a ready-to-use workflow name!


## Minimal example (Prefect vs. Simmate)

It's useful to know how Prefect workflows compare to Simmate workflows. For
simple cases where you have python code, you'd define a Prefect workflow
like so:

``` python
from prefect import flow

@flow
def my_favorite_workflow():
    print("This workflow doesn't do much")
    return 42

# and then run your workflow
state = my_favorite_workflow()
result = state.result()
```

To convert this to a Simmate workflow, we just need to change the format
a little. Instead of a `@flow` decorator, we use the `run_config` method
of a new subclass:

``` python
# NOTE: this example does not follow Simmate's naming convention, so
# some higher level features will be broken. We will fix in a later step.

from simmate.workflow_engine import Workflow

class Example__Python__MyFavoriteSettings(Workflow):

    @staticmethod
    def run_config():
        print("This workflow doesn't do much")
        return 42

# and then run your workflow
state = MyFavoriteWorkflow.run()
result = state.result()
```

Behind the scenes, the `run` method is converting our `run_config` to a
Prefect workflow for us.


### Full example

Now let's look at a realistic example where we build a Workflow that has
input parameters and accesses class attributes/methods:

``` python

class Example__Python__MyFavoriteSettings(Workflow):

    example_constant = 12

    @staticmethod
    def squared(x):
        return x ** 2

    @classmethod
    def run_config(cls, name, say_hello=True, **kwargs):

        if say_hello:
            print(f"Hello and welcome, {name}!")

        # grab class values and methods
        x = cls.example_constant
        example_calc = cls.squared(x)
        print(f"Our calculation gave a result of {example_calc}")

        # grab extra arguments if you need them
        for key, value in kwargs.items():
            print(
                f"An extra parameter for {key} was given "
                "with a value of {value}"
            )

        return "Success!"
```

## Single S3Task + DatabaseTable

In many cases, you may have a workflow that runs a single calculation
and then writes the results to a specific database table. An example of
this would be an energy calculation using VASP. For common tasks
like this, the Workflow class is pre-configured to take an `S3Task`
and `DatabaseTable` -- including a default `run_config` method for you.
You can build your workflow like so:
    
``` python
from simmate.workflow_engine import Workflow

from my_app.tasks import ExampleS3Task
from my_app.models import ExampleTable

class Example__Python__MyFavoriteSettings(Workflow):
    s3task = ExampleS3Task
    database_table = ExampleTable

# try your workflow!
state = Example__Python__MyFavoriteSettings.run()
result = state.result()
```

To learn more about making the S3Task and DatabaseTable, make sure you
have gone through all the Simmate 
[tutorials](https://github.com/jacksund/simmate/tree/main/tutorials), 
including the advanced "Creating custom workflows" tutorial.
"""

import json
import cloudpickle
import yaml
import re
from typing import List
from functools import cache  # cached_property doesnt work with classmethod

from prefect.tasks import task  # present only for convience imports elsewhere
from prefect.flows import Flow
from prefect.states import State
from prefect.context import FlowRunContext
from prefect.client import get_client
from prefect.orion.schemas.filters import FlowFilter, FlowRunFilter
from prefect.packaging import OrionPackager
from prefect.packaging.serializers import PickleSerializer

import simmate
from simmate.toolkit import Structure
from simmate.database.base_data_types import Calculation
from simmate.workflow_engine import S3Task
from simmate.utilities import async_to_sync


class Workflow:
    """
    An abstract base class for all Simmate workflows. Default methods are configured
    for use with S3Tasks and Workflows.
    """

    # TODO: set storage attribute to module

    # TODO: inherit doc from s3task
    # by default we just copy the docstring of the S3task to the workflow
    # workflow.__doc__ = s3task.__doc__

    version: str = simmate.__version__
    """
    Version number for this flow. Defaults to the Simmate version 
    (e.g. "0.7.0").
    """

    s3task: S3Task = None
    """
    The supervised-staged-shell task (or S3Task) that this workflow uses to run.
    For understanding what the calculation does and the settings it uses, users
    should start here. You can also use a workflows `s3task.run` to run the workflow
    without storing results in the database.
    """

    database_table: Calculation = None
    """
    The database table where calculation information (such as the prefect_flow_run_id)
    is stored. The table should use `simmate.database.base_data_types.Calculation`
    
    In many cases, this table will contain all of the results you need. However,
    pay special attention to NestedWorkflows, where your results are often tied
    to a final task.
    """

    description_doc_short: str = None
    """
    A quick description for this workflow. This will be shown in the website UI
    in the list-view of all different workflow presets.
    """

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: str = None,
        copy_previous_directory: bool = False,
        pre_sanitize_structure: bool = None,
        pre_standardize_structure: bool = None,
        is_restart: bool = False,
    ):
        """
        The workflow method, which can be overwritten when inheriting from this
        class. This can be either a staticmethod or classmethod.

        #### The default run method

        There is a default run method implemented that is for S3Tasks.

        Builds a workflow from a S3Task and it's corresponding database table.

        Very often with Simmate's S3Tasks, the workflow for a single S3Task is
        the same. The workflow is typically made of three tasks:

        1. loading the input parameters and registering the calculation
        2. running the calculation (what this S3Task does on its own)
        3. saving the calculation results

        Task 1 and 3 always use the same functions, where we just need to tell
        it which database table we are registering/saving to.

        Because of this common recipe for workflows, we use this method to make
        the workflow for us.
        """
        # local import to prevent circular import error
        from simmate.workflow_engine.common_tasks import (
            load_input_and_register,
            save_result,
        )

        # make sure the workflow is configured properly first
        if not cls.s3task:
            raise NotImplementedError(
                "Please either set the s3task attribute or write a custom run method!"
            )

        parameters_cleaned = load_input_and_register(
            structure=structure,
            command=command,
            source=source,
            directory=directory,
            copy_previous_directory=copy_previous_directory,
            pre_sanitize_structure=pre_sanitize_structure,
            pre_standardize_structure=pre_standardize_structure,
            is_restart=is_restart,
        )

        result = cls.s3task.run(**parameters_cleaned).result()

        save_result(result)

        return result

    @classmethod
    @cache
    def to_prefect_flow(cls) -> Flow:
        """
        Converts this workflow into a Prefect flow
        """

        # Instead of the @flow decorator, we build the flow instance directly
        flow = Flow(
            fn=cls.run_config,
            name=cls.name_full,
            version=cls.version,
            # Skip type checking because I don't have robust typing yet
            # e.g. Structure type inputs also accept inputs like a filename
            validate_parameters=False,
        )

        # as an extra, we set this attribute to the prefect flow instance, which
        # allows us to access the source Simmate Workflow easily with Prefect's
        # context managers.
        flow.simmate_workflow = cls

        return flow

    @classmethod
    def run(cls, **kwargs) -> State:
        """
        A convience method to run a workflow as a subflow in a prefect context.
        """
        subflow = cls.to_prefect_flow()
        state = subflow(return_state=True, **kwargs)

        # We don't want to block and wait because this might disable parallel
        # features of subflows. We therefore return the state and let the
        # user decide if/when to block.
        # result = state.result()

        return state

    @classmethod
    @property
    def name_full(cls) -> str:
        """
        Standardized name of the workflow. This converts the class name like so:
        `Static_Energy__VASP__Matproj` --> `static-energy.vasp.matproj`
        """
        if not len(cls.__name__.split("__")) == 3:
            raise Exception("Make sure you are following Simmate naming conventions!")

        # convert to dot format
        name = cls.__name__.replace("__", ".")

        # adds a hyphen between each capital letter
        # copied from https://stackoverflow.com/questions/199059/
        name = re.sub(r"(\w)([A-Z])", r"\1-\2", name)

        return name.lower()

    @classmethod
    @property
    def name_project(cls) -> str:
        """
        Name of the Project this workflow is associated with. This is the first
        portion of the flow name (e.g. "static-energy")
        """
        return cls.name_full.split(".")[0]

    @classmethod
    @property
    def name_calculator(cls) -> str:
        """
        Name of the calculator this workflow is associated with. This is the second
        portion of the flow name (e.g. "vasp")
        """
        return cls.name_full.split(".")[1]

    @classmethod
    @property
    def name_preset(cls) -> str:
        """
        Name of the settings/preset this workflow is associated with. This is the third
        portion of the flow name (e.g. "matproj" or "matproj-prebader")
        """
        return cls.name_full.split(".")[2]

    # BUG: naming this `description` causes issues.
    # See https://github.com/PrefectHQ/prefect/issues/3911
    @classmethod
    @property
    def description_doc(cls) -> str:
        """
        This simply returns the documentation string of this workflow -- so this
        is the same as `__doc__`. This attribute is only defined for beginners
        to python and for use in django templates for the website interface.
        """
        return cls.__doc__ or cls.s3task.__doc__  # NEEDS REFACTOR

    @classmethod
    @property
    def parameter_names(cls) -> List[str]:
        """
        Gives a list of all the parameter names for this workflow.
        """
        # Iterate through and grab the parameters for the run method. We also
        # sort them alphabetically for consistent results.
        parameter_names = list(cls.to_prefect_flow().parameters.properties.keys())
        parameter_names.sort()
        return parameter_names

    @classmethod
    @property
    def parameter_names_required(cls) -> List[str]:
        """
        Gives a list of all the required parameter names for this workflow.
        """
        return cls.to_prefect_flow().parameters.required

    @classmethod
    def show_parameters(cls):
        """
        Prints a list of all the parameter names for this workflow.
        """
        # use yaml to make the printout pretty (no quotes and separate lines)
        print(yaml.dump(cls.parameter_names))

    @classmethod
    @property
    def parameters_to_register(cls) -> List[str]:
        """
        (experimental feature)
        A list of input parameters that should be used to register the calculation.
        """
        parameters_to_register = [
            "prefect_flow_run_id"
        ]  # run is always used to register but is never an input parameter

        table_columns = cls.database_table.get_column_names()

        for parameter in cls.parameter_names:
            if parameter in table_columns:
                parameters_to_register.append(parameter)

        # check special cases where input parameter doesn't match to a column name
        if "structure_string" in table_columns:
            parameters_to_register.append("structure")

        # put in alphabetical order for consistent results
        parameters_to_register.sort()

        return parameters_to_register

    @classmethod
    def _register_calculation(cls, **kwargs) -> Calculation:
        """
        If the workflow is linked to a calculation table in the Simmate database,
        this adds the flow run to the database.

        Parameters passed should be deserialized and cleaned.

        This method should not be called directly as it is used within the
        `run_cloud` method and `load_input_and_register` task.
        """

        # We first need to grab the database table where we want to register
        # the calculation run to. We can grab the table from either...
        #   1. the database_table attribute
        #   2. flow_context --> flow_name --> flow --> then grab its database_table

        # If this method is being called on the base Workflow class, that
        # means we are trying to register a calculation from within a flow
        # context -- where the context has information such as the workflow
        # we are using (and the database table linked to that workflow).
        if cls == Workflow:

            run_context = FlowRunContext.get()
            workflow = run_context.flow.simmate_workflow
            database_table = workflow.database_table

            # as an extra, add the prefect_flow_run_id the kwargs in case it
            # wasn't set already.
            if "prefect_flow_run_id" not in kwargs:
                kwargs["prefect_flow_run_id"] = str(run_context.flow_run.id)

        # Otherwise we should be using the subclass Workflow that has the
        # database_table property set.
        else:
            workflow = cls  # we have the workflow class already
            database_table = cls.database_table

        # Registration is only possible if a table is provided. Some
        # special-case workflows don't store calculation information bc the flow
        # is just a quick python analysis.
        if not database_table:
            print("No database table found. Skipping registration.")
            return

        # grab the registration kwargs from the parameters provided and then
        # convert them to a python object format for the database method
        register_kwargs = {
            key: kwargs.get(key, None) for key in workflow.parameters_to_register
        }
        register_kwargs_cleaned = cls._deserialize_parameters(**register_kwargs)

        # SPECIAL CASE: for customized workflows we need to convert the inputs
        # back to json before saving to the database.
        if "workflow_base" in register_kwargs_cleaned:
            parameters_serialized = cls._serialize_parameters(**register_kwargs_cleaned)
            prefect_flow_run_id = parameters_serialized.pop("prefect_flow_run_id", None)
            calculation = database_table.from_prefect_id(
                prefect_flow_run_id=prefect_flow_run_id,
                **parameters_serialized,
            )
        else:
            # load/create the calculation for this workflow run
            calculation = database_table.from_prefect_id(**register_kwargs_cleaned)

        return calculation

    @staticmethod
    def _serialize_parameters(**parameters) -> dict:
        """
        Converts input parameters to json-sealiziable objects that Prefect can
        use.

        This method should not be called directly as it is used within the
        run_cloud() method.
        """

        # TODO: consider moving this into prefect's core code as a contribution.
        # This alternatively might be a pydantic contribution

        # Because many flows allow object-type inputs (such as structure object),
        # we need to serialize these inputs before scheduling them with prefect
        # cloud. To do this, I only check for two potential methods:
        #     as_dict
        #     to_dict
        # These were chosen based on common pymatgen methods, but I may change
        # this in the future. As another alternative, I could also cloudpickle
        # input objects when they are not JSON serializable.
        # OPTIMIZE: Prefect current tries to JSON serialize within their
        # client.create_flow_run method. In the future, we may want to move
        # this functionality there.

        parameters_serialized = {}
        for parameter_key, parameter_value in parameters.items():

            try:
                json.dumps(parameter_value)
            except TypeError:
                if hasattr(parameter_value, "as_dict"):
                    parameter_value = parameter_value.as_dict()
                elif hasattr(parameter_value, "to_dict"):
                    parameter_value = parameter_value.to_dict()

                # workflow_base and input_parameters are special cases that
                # may require a refactor (for customized workflows)
                elif parameter_key == "workflow_base":
                    parameter_value = parameter_value.name_full
                elif parameter_key == "input_parameters":
                    # recursive call to this function
                    parameter_value = Workflow._serialize_parameters(**parameter_value)

                else:
                    parameter_value = cloudpickle.dumps(parameter_value)
            parameters_serialized[parameter_key] = parameter_value
        return parameters_serialized

    @staticmethod
    def _deserialize_parameters(**parameters) -> dict:
        """
        converts all parameters to appropriate python objects
        """

        from simmate.toolkit import Structure
        from simmate.toolkit.diffusion import MigrationHop, MigrationImages

        parameters_cleaned = parameters.copy()

        #######
        # SPECIAL CASE: customized workflows have their parameters stored under
        # "input_parameters" instead of the base dict
        # THIS INVOLVES A RECURSIVE CALL TO THIS SAME METHOD
        if "workflow_base" in parameters.keys():
            # This is a non-modular import that can cause issues and slower
            # run times. We therefore import lazily.
            from simmate.workflows.utilities import get_workflow

            # Make sure we have a workflow object
            parameters_cleaned["workflow_base"] = (
                get_workflow(parameters["workflow_base"])
                if isinstance(parameters["workflow_base"], str)
                else parameters["workflow_base"]
            )
            # Make a recursive call for the input parameters
            parameters_cleaned["input_parameters"] = Workflow._deserialize_parameters(
                **parameters["input_parameters"]
            )
            return parameters_cleaned
        #######

        # we don't want to pass arguments like command=None or structure=None if the
        # user didn't provide this input parameter. Instead, we want the workflow to
        # use its own default value. To do this, we first check if the parameter
        # is set in our kwargs dictionary and making sure the value is NOT None.
        # If it is None, then we remove it from our final list of kwargs. This
        # is only done for command, directory, and structure inputs -- as these
        # are the three that are typically assumed to be present (see the CLI).
        parameter_to_filter = [
            "command",
            "directory",
            "pre_sanitize_structure",
            "pre_standardize_structure",
        ]
        for parameter in parameter_to_filter:
            if not parameters.get(parameter, None):
                parameters_cleaned.pop(parameter, None)

        # The remaining checks look to intialize input to toolkit objects

        structure = parameters.get("structure", None)
        if structure:
            parameters_cleaned["structure"] = Structure.from_dynamic(structure)
        else:
            parameters_cleaned.pop("structure", None)

        if "structures" in parameters.keys():
            structure_filenames = parameters["structures"].split(";")
            parameters_cleaned["structures"] = [
                Structure.from_dynamic(file) for file in structure_filenames
            ]

        if "migration_hop" in parameters.keys():
            migration_hop = MigrationHop.from_dynamic(parameters["migration_hop"])
            parameters_cleaned["migration_hop"] = migration_hop

        if "migration_images" in parameters.keys():
            migration_images = MigrationImages.from_dynamic(
                parameters["migration_images"]
            )
            parameters_cleaned["migration_images"] = migration_images

        if "supercell_start" in parameters.keys():
            parameters_cleaned["supercell_start"] = Structure.from_dynamic(
                parameters["supercell_start"]
            )

        if "supercell_end" in parameters.keys():
            parameters_cleaned["supercell_end"] = Structure.from_dynamic(
                parameters["supercell_end"]
            )

        return parameters_cleaned

    # -------------------------------------------------------------------------
    #
    # All methods beyond this point require a Prefect server to be running
    # and Simmate to be connected to it.
    #
    # -------------------------------------------------------------------------

    @classmethod
    @property
    @cache
    @async_to_sync
    async def deployment_id(cls) -> str:
        """
        Grabs the deployment id from the prefect database if it exists, and
        if not, creates the depolyment and then returns the id.

        This is a synchronous and cached version of `_get_deployment_id` and
        this is the preferred method to use for beginners.
        """
        return await cls._get_deployment_id()

    @classmethod
    async def _get_deployment_id(cls) -> str:
        """
        Grabs the deployment id from the prefect database if it exists, and
        if not, creates the depolyment and then returns the id.

        This is an asynchronous method and should only be used when within
        other async methods. Beginners should instead use the `deployment_id`
        property.
        """

        async with get_client() as client:
            response = await client.read_deployments(
                flow_filter=FlowFilter(
                    name={"any_": [cls.name_full]},
                ),
            )

        # If this is the first time accessing the deployment id, we will need
        # to create the deployment
        if not response:
            deployment_id = await cls._create_deployment()

        # there should only be one deployment associated with this workflow
        # if it's been deployed already.
        elif len(response) == 1:
            deployment_id = str(response[0].id)

        else:
            raise Exception("There are duplicate deployments for this workflow!")

        return deployment_id

    @classmethod
    async def _create_deployment(cls) -> str:
        """
        Registers this workflow to the prefect database as a deployment.

        This method should not be called directly. It will be called by
        other methods when appropriate
        """

        # raise error until python-deployments are supported again
        raise Exception(
            "Prefect 2.0 has removed the ability to create deployments in "
            "python, so this feature is currently disabled."
        )
        # When this is removed, be sure to re-add the test_workflow_cloud unittest

        from prefect.deployments import Deployment

        # NOTE: we do not use the client.create_deployment method because it
        # is called within the Deployment.create() method for us.
        deployment = Deployment(
            name=cls.name_full,
            flow=cls.to_prefect_flow(),
            packager=OrionPackager(serializer=PickleSerializer()),
            # OPTIMIZE: it would be better if I could figure out the ImportSerializer
            # here. Only issue is that prefect would need to know to import AND
            # call a method.
            tags=[
                "simmate",
                cls.name_project,
                cls.name_calculator,
            ],
        )

        deployment_id = await deployment.create()

        return str(deployment_id)  # convert from UUID to str first

    @classmethod
    @async_to_sync
    async def _submit_to_prefect(cls, **kwargs) -> str:
        """
        Submits a flow run to prefect cloud.

        This method should not be used directly. Instead use `run_cloud`.
        """

        # The reason we have this code as a separate method is because we want
        # to isolate Prefect's async calls from Django's sync-restricted calls
        # (i.e. django raises errors if called within an async context).
        # Therefore, methods like `run_cloud` can't have both this async code
        # AND methods like _register_calculation that make sync database calls.

        async with get_client() as client:
            response = await client.create_flow_run_from_deployment(
                deployment_id=await cls._get_deployment_id(),
                **kwargs,
            )

        flow_run_id = str(response.id)
        return flow_run_id

    @classmethod
    def run_cloud(cls, **kwargs) -> str:
        """
        This schedules the workflow to run remotely on Prefect Cloud.

        #### Parameters

        - `labels`:
            a list of labels to schedule the workflow with

        - `wait_for_run`:
            whether to wait for the workflow to finish. If False, the workflow
            will simply be submitted and then exit. The default is True.

        - `**kwargs`:
            all options that are normally passed to the workflow.run() method

        #### Returns

        - The flow run id that was used in prefect cloud.


        #### Usage

        Make sure you have Prefect properly configured and have registered your
        workflow with the backend.

        Note that this method can be viewed as a fork of:
            - from prefect.tasks.prefect.flow_run import create_flow_run
        It can also be viewed as a more convenient way to call to client.create_flow_run.
        I don't accept any other client.create_flow_run() inputs besides 'labels'.
        This may change in the future if I need to set flow run names or schedules.
        """

        # Prefect does not properly deserialize objects that have
        # as as_dict or to_dict method, so we use a custom method to do that here
        parameters_serialized = cls._serialize_parameters(**kwargs)
        # BUG: What if we are submitting using a filename? We don't want to
        # submit to a cluster and have the job fail because it doesn't have
        # access to the file. One solution could be to deserialize right before
        # serializing in the next line in order to ensure parameters that
        # accept file names are submitted with all necessary data.

        # Now we submit the workflow.
        flow_run_id = cls._submit_to_prefect(parameters=parameters_serialized)

        # Because we often want to save some info to our database even before
        # the calculation starts/finishes, we do that here. An example is
        # storing the structure and prefect id that we just submitted.
        cls._register_calculation(prefect_flow_run_id=flow_run_id, **kwargs)
        # BUG: Will there be a race condition here? What if the workflow finishes
        # and tries writing to the databse before this is done?
        # BUG: if parameters are improperly set, this line will fail, while the
        # job submission (above) will suceed. Should I cancel the flow run if
        # this occurs?

        # return the flow_run_id for the user
        return flow_run_id

    @classmethod
    @property
    @async_to_sync
    async def nflows_submitted(cls) -> int:
        """
        Queries Prefect to see how many workflows are in a scheduled, running,
        or pending state.
        """

        async with get_client() as client:
            response = await client.read_flow_runs(
                flow_filter=FlowFilter(
                    name={"any_": [cls.name_full]},
                ),
                flow_run_filter=FlowRunFilter(
                    state={"type": {"any_": ["SCHEDULED", "PENDING", "RUNNING"]}}
                ),
            )

        return len(response)



# -*- coding: utf-8 -*-

import os
from pathlib import Path

# grab the users set directory for all of their VASP POTCAR files
# TODO - for now, I assume that the directory is located in...
#   [home_directory] ~/simmate/vasp/Potentials
potcar_dir = os.path.join(Path.home(), "simmate", "vasp", "Potentials")


# This maps out where functional POTCARs are located. All of these
# files should be located in the same Potentals directory and follow
# the original folder structure provided by VASP.
FOLDER_MAPPINGS = {
    "LDA": os.path.join(potcar_dir, "LDA", "potpaw_LDA.54"),
    "PBE": os.path.join(potcar_dir, "PBE", "potpaw_PBE.54"),
    "PBE_GW": os.path.join(potcar_dir, "PBE", "potpaw_PBE.54"),
}

# We need to map which POTCAR to grab for each element based off of the type
# of calculation we are doing. To see how these were selected, you should look
# at the following:
#
#   Material's Project rational
#   https://wiki.materialsproject.org/Pseudopotentials_Choice
#
#   Pymatgen sets (look at the yaml files)
#   https://github.com/materialsproject/pymatgen/tree/v2020.12.31/pymatgen/io/vasp
#   For potential that use fewer electron (faster but lower quality), it's useful
#   to look at the MITRelaxSet.yaml, but MPRelaxSet's choices of higher electron
#   counts can give fast calculations (<1min) AND are appropriate for high-quality
#   calculations. Thus we go ahead with the more accurate ones as the default choice.
#
#   VASP recommendations
#   https://cms.mpi.univie.ac.at/vasp/vasp/Ultrasoft_pseudopotentials_supplied_with_VASP_package.html
#   Note, this webpage is outdated, but I don't see them move this information
#   into their new wiki...
#
#   Comparison of different potentials' accuracy
#   https://molmod.ugent.be/deltacodesdft

# TODO -- add LDA mappings. There's no example in Pymatgen that I see.

PBE_ELEMENT_MAPPINGS = {
    "Ac": "Ac",
    "Ag": "Ag",
    "Al": "Al",
    "Ar": "Ar",
    "As": "As",
    "Au": "Au",
    "B": "B",
    "Ba": "Ba_sv",
    "Be": "Be_sv",
    "Bi": "Bi",
    "Br": "Br",
    "C": "C",
    "Ca": "Ca_sv",
    "Cd": "Cd",
    "Ce": "Ce",
    "Cl": "Cl",
    "Co": "Co",
    "Cr": "Cr_pv",
    "Cs": "Cs_sv",
    "Cu": "Cu_pv",
    "Dy": "Dy_3",
    "Er": "Er_3",
    "Eu": "Eu",
    "F": "F",
    "Fe": "Fe_pv",
    "Ga": "Ga_d",
    "Gd": "Gd",
    "Ge": "Ge_d",
    "H": "H",
    "He": "He",
    "Hf": "Hf_pv",
    "Hg": "Hg",
    "Ho": "Ho_3",
    "I": "I",
    "In": "In_d",
    "Ir": "Ir",
    "K": "K_sv",
    "Kr": "Kr",
    "La": "La",
    "Li": "Li_sv",
    "Lu": "Lu_3",
    "Mg": "Mg_pv",
    "Mn": "Mn_pv",
    "Mo": "Mo_pv",
    "N": "N",
    "Na": "Na_pv",
    "Nb": "Nb_pv",
    "Nd": "Nd_3",
    "Ne": "Ne",
    "Ni": "Ni_pv",
    "Np": "Np",
    "O": "O",
    "Os": "Os_pv",
    "P": "P",
    "Pa": "Pa",
    "Pb": "Pb_d",
    "Pd": "Pd",
    "Pm": "Pm_3",
    "Pr": "Pr_3",
    "Pt": "Pt",
    "Pu": "Pu",
    "Rb": "Rb_sv",
    "Re": "Re_pv",
    "Rh": "Rh_pv",
    "Ru": "Ru_pv",
    "S": "S",
    "Sb": "Sb",
    "Sc": "Sc_sv",
    "Se": "Se",
    "Si": "Si",
    "Sm": "Sm_3",
    "Sn": "Sn_d",
    "Sr": "Sr_sv",
    "Ta": "Ta_pv",
    "Tb": "Tb_3",
    "Tc": "Tc_pv",
    "Te": "Te",
    "Th": "Th",
    "Ti": "Ti_pv",
    "Tl": "Tl_d",
    "Tm": "Tm_3",
    "U": "U",
    "V": "V_pv",
    "W": "W_pv",
    "Xe": "Xe",
    "Y": "Y_sv",
    "Yb": "Yb_2",
    "Zn": "Zn",
    "Zr": "Zr_sv",
}

# Another common mapping comes from the MIT project (pre-cursor to Materials Project)
# The POTCARs here are generally the ones with fewer electrons and convergence
# criteria is less tight. A lot of the elements are the same as above, so we
# just copy those element mappings and update the elements that are actaully
# lower quality here.
PBE_ELEMENT_MAPPINGS_LOW_QUALITY = PBE_ELEMENT_MAPPINGS.copy()
PBE_ELEMENT_MAPPINGS_LOW_QUALITY.update(
    {
        "Be": "Be",
        "Cr": "Cr",
        "Cu": "Cu",
        "Fe": "Fe",
        "Ga": "Ga",
        "Ge": "Ge",
        "Hf": "Hf",
        "In": "In",
        "Li": "Li",
        "Mg": "Mg",
        "Mn": "Mn",
        "Na": "Na",
        "Nd": "Nd",
        "Ni": "Ni",
        "Os": "Os",
        "Pb": "Pb",
        "Pm": "Pm",
        "Pr": "Pr",
        "Rb": "Rb_pv",
        "Re": "Re",
        "Rh": "Rh",
        "Ta": "Ta",
        "Tc": "Tc",
        "Ti": "Ti",
        "Tl": "Tl",
        "Yb": "Yb",
        "Zr": "Zr",
    }
)

PBE_GW_ELEMENT_MAPPINGS = {
    "Ac": "Ac",
    "Ag": "Ag_sv_GW",
    "Al": "Al_GW",
    "Ar": "Ar_GW",
    "As": "As_GW",
    "At": "At_d_GW",
    "Au": "Au_sv_GW",
    "B": "B_GW",
    "Ba": "Ba_sv_GW",
    "Be": "Be_sv_GW",
    "Bi": "Bi_d_GW",
    "Br": "Br_GW",
    "C": "C_GW",
    "Ca": "Ca_sv_GW",
    "Cd": "Cd_sv_GW",
    "Ce": "Ce_GW",
    "Cl": "Cl_GW",
    "Co": "Co_sv_GW",
    "Cr": "Cr_sv_GW",
    "Cs": "Cs_sv_GW",
    "Cu": "Cu_sv_GW",
    "Dy": "Dy_3",
    "Er": "Er_3",
    "Eu": "Eu",
    "F": "F_GW",
    "Fe": "Fe_sv_GW",
    "Ga": "Ga_d_GW",
    "Gd": "Gd",
    "Ge": "Ge_d_GW",
    "H": "H_GW",
    "He": "He_GW",
    "Hf": "Hf_sv_GW",
    "Hg": "Hg_sv_GW",
    "Ho": "Ho_3",
    "I": "I_GW",
    "In": "In_d_GW",
    "Ir": "Ir_sv_GW",
    "K": "K_sv_GW",
    "Kr": "Kr_GW",
    "La": "La_GW",
    "Li": "Li_sv_GW",
    "Lu": "Lu_3",
    "Mg": "Mg_sv_GW",
    "Mn": "Mn_sv_GW",
    "Mo": "Mo_sv_GW",
    "N": "N_GW",
    "Na": "Na_sv_GW",
    "Nb": "Nb_sv_GW",
    "Nd": "Nd_3",
    "Ne": "Ne_GW",
    "Ni": "Ni_sv_GW",
    "Np": "Np",
    "O": "O_GW",
    "Os": "Os_sv_GW",
    "P": "P_GW",
    "Pa": "Pa",
    "Pb": "Pb_d_GW",
    "Pd": "Pd_sv_GW",
    "Pm": "Pm_3",
    "Po": "Po_d_GW",
    "Pr": "Pr_3",
    "Pt": "Pt_sv_GW",
    "Pu": "Pu",
    "Rb": "Rb_sv_GW",
    "Re": "Re_sv_GW",
    "Rh": "Rh_sv_GW",
    "Rn": "Rn_d_GW",
    "Ru": "Ru_sv_GW",
    "S": "S_GW",
    "Sb": "Sb_d_GW",
    "Sc": "Sc_sv_GW",
    "Se": "Se_GW",
    "Si": "Si_GW",
    "Sm": "Sm_3",
    "Sn": "Sn_d_GW",
    "Sr": "Sr_sv_GW",
    "Ta": "Ta_sv_GW",
    "Tb": "Tb_3",
    "Tc": "Tc_sv_GW",
    "Te": "Te_GW",
    "Th": "Th",
    "Ti": "Ti_sv_GW",
    "Tl": "Tl_d_GW",
    "Tm": "Tm_3",
    "U": "U",
    "V": "V_sv_GW",
    "W": "W_sv_GW",
    "Xe": "Xe_GW",
    "Y": "Y_sv_GW",
    "Yb": "Yb_2",
    "Zn": "Zn_sv_GW",
    "Zr": "Zr_sv_GW",
}



# -*- coding: utf-8 -*-

"""
Workflows for analyzing ionic diffusion with a material
"""

from simmate.calculators.vasp.workflows.diffusion.all import (
    Diffusion__Vasp__NebAllPaths,
    Diffusion__Vasp__NebSinglePath,
    Diffusion__Vasp__NebFromEndpoints,
    Diffusion__Vasp__NebFromImages,
)



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.database.base_data_types import Spacegroup as SpacegroupTable


class Spacegroup(filters.FilterSet):
    class Meta:
        model = SpacegroupTable
        fields = dict(
            number=["exact", "range"],
            symbol=["exact"],
            crystal_system=["exact"],
            point_group=["exact"],
        )
